from . import utils

class Minimalism:
    """
    Main class for musical minimalism.
    """

    class Process:
        """
        Class for the process of musical minimalism.
        """

        def __init__(self, operation='additive', direction='forward', repetition=0):
            """
            Initializes the process with the specified operation and direction.

            Args:
                operation (str): The operation to be used. Can be 'additive' or 'subtractive'.
                direction (str): The direction of the operation. Can be 'forward', 'backward', 'inward' or 'outward'.
                repetition (int): The number of times the process is repeated.
            """

            if operation in ['additive', 'subtractive']:
                self.operation = operation
            else:
                raise ValueError("Invalid output type. Choose 'additive' or 'subtractive'.")
            if direction in ['forward', 'backward', 'inward', 'outward']:
                self.direction = direction
            else:
                raise ValueError("Invalid output type. Choose 'forward', 'backward', 'inward' or 'outward'.")
            
            if repetition < 0 or not isinstance(repetition, int):
                raise ValueError("Invalid repetition value. Must be an integer greater of equal to 0.")
            self.repetition = repetition
            
        # Additive operations
        # -------------------
        def additive_forward(self):
            """
            Applies the additive forward operation to the sequence with repetition, as::

                [
                    'C4',
                    'C4', 'D4',
                    'C4', 'D4', 'E4',
                    'C4', 'D4', 'E4', 'F4',
                    'C4', 'D4', 'E4', 'F4', 'G4',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5'
                ]

            Returns:
                list: The processed sequence.
            """
            processed = []
            for i in range(len(self.sequence)):
                segment = [self.sequence[j] for j in range(i + 1)]
                for _ in range(self.repetition+1):
                    processed.extend(segment)
            return processed
        
        def additive_backward(self):
            """
            Applies the additive backward operation to the sequence with repetition, as::

                [
                    'C5',
                    'B4', 'C5',
                    'A4', 'B4', 'C5',
                    'G4', 'A4', 'B4', 'C5',
                    'F4', 'G4', 'A4', 'B4', 'C5',
                    'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                ]

            Returns:
                list: The processed sequence.
            """
            processed = []
            for i in range(len(self.sequence), 0, -1): 
                segment = [self.sequence[j] for j in range(i - 1, len(self.sequence))]
                for _ in range(self.repetition+1): 
                    processed.extend(segment)
            return processed

        def additive_inward(self):
            """
            Applies the additive inward operation to the sequence, as::

                [
                    'C4',                                     'C5',
                    'C4', 'D4',                         'B4', 'C5',
                    'C4', 'D4', 'E4',             'A4', 'B4', 'C5',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5'
                ]

            Returns:
                list: The processed sequence.
            """

            n = len(self.sequence)
            processed = []
            segment = []
            for i in range((n + 1) // 2):  # Ensuring we cover all elements
                if i < n - i - 1:  # Avoid adding the middle element twice
                    segment = [self.sequence[:i+1], self.sequence[(n - i-1):]]
                else:
                    segment = [self.sequence]
                for _ in range(self.repetition+1):
                    processed.extend(segment)
            
            processed = [item for sublist in processed for item in sublist]
            return processed
        
        def additive_outward(self):
            """
            Applies the additive outward operation to the sequence with repetition, as::

                [
                                    'F4', 'G4',
                                'E4', 'F4', 'G4', 'A4',
                        'D4', 'E4', 'F4', 'G4', 'A4', 'B4',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5'
                ]

            Returns:
                list: The processed sequence.
            """

            n = len(self.sequence)
            processed = []

            # Check if the sequence is odd or even and find the middle
            if n % 2 == 0:  # Even
                mid_left = n // 2 - 1
                mid_right = n // 2
                # Start from the middle two elements and expand outwards
                for i in range(n // 2):
                    segment = self.sequence[mid_left - i : mid_right + i + 1]
                    for _ in range(self.repetition+1):
                        processed.extend(segment)
            else:  # Odd
                mid = n // 2
                # Start from the middle element and expand outwards
                for i in range(mid + 1):
                    segment = self.sequence[mid - i : mid + i + 1]
                    for _ in range(self.repetition+1):
                        processed.extend(segment)

            return processed
        
        # Subtractive operations
        # ----------------------
        def subtractive_forward(self):
            """
            Applies the subtractive forward operation to the sequence, as::

                [
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'F4', 'G4', 'A4', 'B4', 'C5',
                    'G4', 'A4', 'B4', 'C5',
                    'A4', 'B4', 'C5',
                    'B4', 'C5',
                    'C5'
                ]

            Returns:
                list: The processed sequence.
            """

            processed = []
            for i in range(len(self.sequence)):
                # Each time, remove one more note from the beginning
                segment = self.sequence[i:]
                for _ in range(self.repetition+1):  # Here we consider if repetition is needed
                    processed.extend(segment)  # Directly append the segment without extending it, because we're dealing with segments now
            return processed
        
        def subtractive_backward(self):
            """
            Applies the subtractive backward operation to the sequence, as::

                [
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4',
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4',
                    'C4', 'D4', 'E4', 'F4', 'G4',
                    'C4', 'D4', 'E4', 'F4',
                    'C4', 'D4', 'E4',
                    'C4', 'D4',
                    'C4'
                ]

            Returns:
                list: The processed sequence.
            """

            processed = []
            for i in range(len(self.sequence), 0, -1):  # Start from the full sequence and decrement
                # Each time, create a segment that removes one more note from the end
                segment = self.sequence[:i]
                for _ in range(self.repetition+1):  # Consider if repetition is needed
                    processed.extend(segment)  # Append the segment directly
            return processed
        
        def subtractive_inward(self):
            """
            Applies the subtractive inward operation to the sequence, as::

                [
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                        'D4', 'E4', 'F4', 'G4', 'A4', 'B4',
                                'E4', 'F4', 'G4', 'A4',
                                    'F4', 'G4',
                ]

            Returns:
                list: The processed sequence.
            """

            n = len(self.sequence)
            steps = n // 2
            # Compute the steps for subtraction, which will be half the length of the sequence rounded down
            processed = self.sequence * (self.repetition + 1)

            # Add segments, removing elements from both ends, also repeated
            for i in range(1, steps + 1):
                segment = self.sequence[i: n - i]  # Remove elements from both ends
                if segment:  # Check if there's anything left to add.
                    processed += segment * (self.repetition + 1)
            return processed
        
        def subtractive_outward(self):
            """
            Applies the subtractive outward operation to the sequence, as::

                [
                    'C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4', 'C5',
                    'C4', 'D4', 'E4',           , 'A4', 'B4', 'C5',
                    'C4', 'D4',                         'B4', 'C5',
                    'C4',                                     'C5',
                ]

            Returns:
                list: The processed sequence.
            """

            segment = self.sequence
            processed = segment * (self.repetition+1)
            while len(segment) > 2:
                segment = segment[1:-1]  # remove the first and last elements
                for _ in range(self.repetition+1):
                    processed.extend(segment)
            return processed

        
        def generate(self, sequence):
            """
            Generates the processed sequence based on the operation and direction.

            Args:
                sequence (list): The sequence to be processed.

            Returns:
                list: The processed sequence.
            """
            self.sequence = sequence
            
            if self.operation == 'additive' and self.direction == 'forward':
                processed = self.additive_forward()
            elif self.operation == 'additive' and self.direction == 'backward':
                processed = self.additive_backward()
            elif self.operation == 'additive' and self.direction == 'inward':
                processed = self.additive_inward()
            elif self.operation == 'additive' and self.direction == 'outward':
                processed = self.additive_outward()
            elif self.operation == 'subtractive' and self.direction == 'forward':
                processed = self.subtractive_forward()
            elif self.operation == 'subtractive' and self.direction == 'backward':
                processed = self.subtractive_forward()
            elif self.operation == 'subtractive' and self.direction == 'inward':
                processed = self.subtractive_inward()
            elif self.operation == 'subtractive' and self.direction == 'outward':
                processed = self.subtractive_outward()

            new_processed = []
            current_offset = 0
            
            # Adjust offsets based on the duration
            for note in processed:
                new_note = (note[0], note[1], current_offset)
                new_processed.append(new_note)
                current_offset += note[1]

            return new_processed

    class Tintinnabuli:
        """
        Class for the Tintinnabuli style of musical minimalism.
        """

        def __init__(self, t_chord, direction='down', rank=0):
            """
            Initializes the Tintinnabuli style with the specified t-chord, direction, and rank.

            Args:
                t_chord (list): The t-chord to be used.
                direction (str): The direction of the operation. Can be 'up', 'down', 'any' or 'alternate'.
                rank (int): The rank of the t-chord.
            """

            if direction in ['up', 'down', 'any', 'alternate']:
                self.direction = direction
            else:
                raise ValueError("Invalid output type. Choose 'up', 'down', 'any' or 'alternate'.")
            
            if self.direction == 'alternate':
                self.is_alternate = True
                self.direction = 'up'
            else:
                self.is_alternate = False
            
            self.t_chord = t_chord
            if not isinstance(rank, int) or rank < 0:
                raise ValueError("Rank must be a non-negative integer lower or equal to the length of the t-chord.")
            else:
                self.rank = rank
            
            if self.rank >= len(self.t_chord):
                self.rank = len(self.t_chord) - 1
                print("Rank exceeds the length of the t-chord. Defaulting to the last note of the t-chord.")

        def generate(self, sequence):
            """
            Generates the t-voice based on the t-chord, direction, and rank.

            Args:
                sequence (list): The m-voice sequence.

            Returns:
                list: The t-voice sequence.
            """
            
            self.sequence = sequence  # the m-voice
            if utils.check_input(self.sequence) == 'list of tuples':
                pitch_sequence = [note[0] for note in self.sequence]
            elif utils.check_input(self.sequence) == 'list':
                pitch_sequence = self.sequence
            else:
                raise ValueError("Invalid input type. Must be a list of tuples or a list.")
            
            t_voice = []
            for m in pitch_sequence:
                if m is None:
                    t_voice.append(None)
                    continue
                differences = [t - m for t in self.t_chord]
                sorted_differences = sorted(enumerate(differences), key=lambda x: abs(x[1]))

                # Ensure rank is within the range of available notes
                effective_rank = self.rank  # Default to the specified rank
                
                if self.direction in ['up', 'down']:
                    filtered_differences = [(index, value) for index, value in sorted_differences if value >= 0] if self.direction == 'up' else [(index, value) for index, value in sorted_differences if value <= 0]
                    if not filtered_differences:
                        # If there are no notes in the desired direction, default to the extreme note in that direction
                        t_voice_i = max(self.t_chord) if self.direction == 'up' else min(self.t_chord)
                    else:
                        # Adjust rank if it exceeds the length of filtered differences
                        if effective_rank >= len(filtered_differences):
                            effective_rank = len(filtered_differences) - 1  # Use the last available note
                        
                        chosen_index = filtered_differences[effective_rank][0]
                        t_voice_i = self.t_chord[chosen_index]
                elif self.direction == 'any':
                    # Adjust rank if it exceeds the length of sorted differences
                    if effective_rank >= len(sorted_differences):
                        effective_rank = len(sorted_differences) - 1  # Use the last available note
                    
                    chosen_index = sorted_differences[effective_rank][0]
                    t_voice_i = self.t_chord[chosen_index]

                # Change direction if alternate
                if self.is_alternate:
                    self.direction = 'down' if self.direction == 'up' else 'up'
                
                t_voice.append(t_voice_i)
            
            if utils.check_input(self.sequence) == 'list of tuples':
                t_voice = [(t_voice[i], note[1], note[2]) for i, note in enumerate(self.sequence)]

            return t_voice



import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# Project information
project = 'djalgo'
copyright = '2024, Essi Parent'
author = 'Essi Parent'
release = "0.2.2"

# Extensions
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'recommonmark',
    'nbsphinx',
]

# General configuration
master_doc = 'index'
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
    '.ipynb': 'jupyter_notebook',
}

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# HTML output configuration
html_theme = 'furo'
html_static_path = ['_static']
templates_path = ['_templates']

# Nbsphinx settings
nbsphinx_execute = 'never'
html_logo = '_static/logo.png'


from pathlib import Path
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from miditok import REMI, TokenizerConfig
from miditok.pytorch_data import DatasetMIDI, DataCollator

class ModelManager:
    """
    Manages the initialization, training, and generation of a neural network model for MIDI sequence generation.

    Attributes:
        seq_len_input (int): Length of input sequences.
        seq_len_output (int): Length of output sequences.
        model_type (str): Type of model to use ('lstm', 'gru', 'transformer').
        nn_units (tuple): Number of units in each layer of the neural network.
        dropout (float): Dropout rate for the neural network.
        batch_size (int): Number of samples per batch.
        learning_rate (float): Learning rate for the optimizer.
        n_heads (int, optional): Number of attention heads (used for transformer model).
        tokenizer (miditok.REMI): Tokenizer for MIDI files.
        input_size (int): Size of the input vocabulary.
        model (torch.nn.Module): Neural network model.
        optimizer (torch.optim.Optimizer): Optimizer for training the model.
        loss_fn (torch.nn.Module): Loss function for training the model.
    """
    def __init__(self, sequence_length_input, sequence_length_output, model_type, nn_units, dropout, batch_size, learning_rate, n_heads=None):
        """
        Initializes the ModelManager with specified parameters.

        Args:
            sequence_length_input (int): Length of input sequences.
            sequence_length_output (int): Length of output sequences.
            model_type (str): Type of model to use ('lstm', 'gru', 'transformer').
            nn_units (tuple): Number of units in each layer of the neural network.
            dropout (float): Dropout rate for the neural network.
            batch_size (int): Number of samples per batch.
            learning_rate (float): Learning rate for the optimizer.
            n_heads (int, optional): Number of attention heads (used for transformer model).
        """
        self.seq_len_input = sequence_length_input
        self.seq_len_output = sequence_length_output
        self.model_type = model_type
        self.nn_units = nn_units
        self.dropout = dropout
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.n_heads = n_heads

        # Initialize the MidiTok tokenizer with custom configuration
        TOKENIZER_PARAMS = {
            "pitch_range": (21, 109),
            "beat_res": {(0, 4): 8, (4, 12): 4},
            "num_velocities": 32,
            "special_tokens": ["PAD", "BOS", "EOS", "MASK"],
            "use_chords": True,
            "use_rests": True,
            "use_tempos": True,
            "use_time_signatures": False,
            "use_programs": False,
            "num_tempos": 32,
            "tempo_range": (40, 250),
        }
        config = TokenizerConfig(**TOKENIZER_PARAMS)
        self.tokenizer = REMI(config)
        self.input_size = len(self.tokenizer.vocab)

        # Create the model
        if model_type == 'transformer':
            self.model = TransformerModel(nn_units, dropout, n_heads, self.input_size)
        elif model_type == 'lstm':
            self.model = LSTMModel(nn_units, dropout, self.input_size)
        elif model_type == 'gru':
            self.model = GRUModel(nn_units, dropout, self.input_size)
        else:
            raise ValueError("Unsupported model type. Choose 'lstm', 'gru', or 'transformer'.")

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
        self.loss_fn = nn.CrossEntropyLoss()

    def fit(self, midi_files_path, epochs, verbose=None):
        """
        Trains the model on the specified MIDI files for a given number of epochs.

        Args:
            midi_files_path (str): Path to the directory containing MIDI files.
            epochs (int): Number of epochs to train the model.
            verbose (int or None, optional): Number of steps between progress messages or None for no messages (default: None).
        
        Returns:
            torch.nn.Module: Trained model.
        """
        dataset_chunks_dir = Path(midi_files_path)
        dataset = DatasetMIDI(
            files_paths=list(dataset_chunks_dir.glob("**/*.mid")),
            tokenizer=self.tokenizer,
            max_seq_len=self.seq_len_input
        )
        collator = DataCollator(self.tokenizer.pad_token_id, copy_inputs_as_labels=True)
        data_loader = DataLoader(dataset, batch_size=self.batch_size, collate_fn=collator)

        self.model.train()
        current_step = 0
        for epoch in range(epochs):
            for batch in data_loader:
                inputs, targets = batch['input_ids'], batch['labels']

                # One-hot encode inputs
                inputs_one_hot = torch.nn.functional.one_hot(inputs, num_classes=self.input_size).float()
                if torch.isnan(inputs_one_hot).any() or torch.isinf(inputs_one_hot).any():
                    print("Input contains NaN or Inf values.")

                if self.model_type == 'transformer':
                    attention_mask = batch['attention_mask']
                    attention_mask_bool = attention_mask.bool()
                    outputs = self.model(inputs_one_hot, attention_mask=attention_mask_bool)
                else:
                    outputs = self.model(inputs_one_hot)  # No attention mask for other types

                loss = self.loss_fn(outputs.view(-1, outputs.size(-1)), targets.view(-1))
                if torch.isnan(loss).any() or torch.isinf(loss).any():
                    print("Loss contains NaN or Inf values.")
                    continue

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                if verbose and current_step % verbose == 0:
                    print(f"Epoch {epoch + 1}/{epochs}, Step {current_step}, Loss: {loss.item()}")

                current_step += 1


        return self.model

    def generate(self, length, primer_file, output_file):
        """
        Generates a MIDI sequence based on a primer file and saves it to an output file.

        Args:
            length (int): Length of the sequence to generate.
            primer_file (str): Path to the primer MIDI file.
            output_file (str): Path to save the generated MIDI file.
        """
        # Tokenize the primer file using the correct method
        midi = self.tokenizer.encode(Path(primer_file))
        primer_tokens = midi[0].ids  # Assuming single-track MIDI
        primer_one_hot = torch.nn.functional.one_hot(torch.tensor(primer_tokens), num_classes=self.input_size).float().unsqueeze(0)

        # Simplified prediction method
        self.model.eval()
        generated = primer_one_hot
        for _ in range(length):
            output = self.model(generated)
            next_token = torch.argmax(output[:, -1, :], dim=-1, keepdim=True)
            next_one_hot = torch.nn.functional.one_hot(next_token, num_classes=self.input_size).float()
            generated = torch.cat((generated, next_one_hot), dim=1)

        # Convert generated tokens back to MIDI
        generated_tokens = torch.argmax(generated, dim=-1).squeeze().tolist()
        generated_midi = self.tokenizer.decode([generated_tokens])
        generated_midi.dump_midi(Path(output_file))

    def save(self, file_path):
        """
        Saves the model state to a file.

        Args:
            file_path (str): Path to save the model state.
        """
        torch.save(self.model.state_dict(), file_path)


class LSTMModel(nn.Module):
    """
    Defines an LSTM model for sequence generation.

    Attributes:
        layers (torch.nn.ModuleList): List of LSTM layers.
        fc (torch.nn.Linear): Fully connected layer for output.
    """
    def __init__(self, nn_units, dropout, input_size):
        """
        Initializes the LSTMModel.

        Args:
            nn_units (tuple): Number of units in each LSTM layer.
            dropout (float): Dropout rate for the LSTM layers.
            input_size (int): Size of the input vocabulary.
        """
        super().__init__()
        self.layers = nn.ModuleList([
            nn.LSTM(input_size=input_size if i == 0 else nn_units[i-1],
                    hidden_size=nn_units[i],
                    batch_first=True,
                    dropout=dropout if i < len(nn_units) - 1 else 0)  # dropout for all but the last layer
            for i in range(len(nn_units))
        ])
        self.fc = nn.Linear(nn_units[-1], input_size)

    def forward(self, x):
        """
        Defines the forward pass of the LSTMModel.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        for layer in self.layers:
            x, _ = layer(x)
        return self.fc(x)


class GRUModel(nn.Module):
    """
    Defines a GRU model for sequence generation.

    Attributes:
        layers (torch.nn.ModuleList): List of GRU layers.
        fc (torch.nn.Linear): Fully connected layer for output.
    """
    def __init__(self, nn_units, dropout, input_size):
        """
        Initializes the GRUModel.

        Args:
            nn_units (tuple): Number of units in each GRU layer.
            dropout (float): Dropout rate for the GRU layers.
            input_size (int): Size of the input vocabulary.
        """
        super().__init__()
        self.layers = nn.ModuleList([
            nn.GRU(input_size=input_size if i == 0 else nn_units[i-1],
                   hidden_size=nn_units[i],
                   batch_first=True,
                   dropout=dropout if i < len(nn_units) - 1 else 0)  # dropout for all but the last layer
            for i in range(len(nn_units))
        ])
        self.fc = nn.Linear(nn_units[-1], input_size)

    def forward(self, x):
        """
        Defines the forward pass of the GRUModel.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        for layer in self.layers:
            x, _ = layer(x)
        return self.fc(x)


class TransformerModel(nn.Module):
    """
    Defines a Transformer model for sequence generation.

    Attributes:
        embedding (torch.nn.Linear): Embedding layer to project input size to embedding dimension.
        encoder_layers (torch.nn.TransformerEncoderLayer): Transformer encoder layer.
        transformer_encoder (torch.nn.TransformerEncoder): Transformer encoder.
        fc (torch.nn.Linear): Fully connected layer for output.
    """
    def __init__(self, nn_units, dropout, n_heads, input_size):
        """
        Initializes the TransformerModel.

        Args:
            nn_units (tuple): Number of units in each transformer layer.
            dropout (float): Dropout rate for the transformer layers.
            n_heads (int): Number of attention heads.
            input_size (int): Size of the input vocabulary.
        """
        super().__init__()
        self.embedding = nn.Linear(input_size, nn_units[0])  # Project input_size to the embedding dimension
        nn.init.xavier_uniform_(self.embedding.weight)  # Xavier initialization
        self.encoder_layers = nn.TransformerEncoderLayer(d_model=nn_units[0], nhead=n_heads, dropout=dropout, batch_first=True, norm_first=True)
        self.transformer_encoder = nn.TransformerEncoder(self.encoder_layers, num_layers=len(nn_units))
        self.fc = nn.Linear(nn_units[-1], input_size)
        nn.init.xavier_uniform_(self.fc.weight)  # Xavier initialization

    def forward(self, x, attention_mask=None):
        """
        Defines the forward pass of the TransformerModel.

        Args:
            x (torch.Tensor): Input tensor.
            attention_mask (torch.Tensor, optional): Attention mask for the transformer.

        Returns:
            torch.Tensor: Output tensor.
        """
        x = self.embedding(x)  # Project input to the embedding dimension
        if attention_mask is not None:
            x = self.transformer_encoder(x, src_key_padding_mask=attention_mask)
        else:
            x = self.transformer_encoder(x)
        return self.fc(x)



import copy
import random
from .analysis import Index

class Darwin:
    def __init__(
            self, initial_phrases, mutation_rate=0.05, population_size=50,
            mutation_probabilities=None,
            scale=None, measure_length = 4, time_resolution=[0.125, 4],
            weights = None, targets=None, seed=None
        ):
        """
        Initialize the Darwin class for evolving musical phrases.

        Args:
            initial_phrases (list): List of initial musical phrases, each a list of tuples (pitch, duration, offset).
            index_weights (dict): Weights for each musical aspect ('gini', 'balance', 'motif', 'dissonance').
            mutation_rate (float): Probability of mutation per attribute in a phrase.
            population_size (int): Number of phrases in each generation.
            scale (list): List of pitches considered harmonious or in-scale; used for dissonance calculation.
        """
        if mutation_probabilities:
            self.mutation_probabilities = mutation_probabilities
        else:
            self.mutation_probabilities = {
                'pitch': lambda: random.choice(self.scale),
                'duration': lambda: random.choice(self.possible_durations),
                'rest': lambda: None if random.random() < 0.02 else 1  # 2% chance of a rest
            }
        self.population = [self.mutate(phrase, rate=0) for phrase in initial_phrases for _ in range(population_size // len(initial_phrases))]
        self.weights = weights
        self.mutation_rate = mutation_rate
        self.population_size = population_size

        self.scale = scale
        self.measure_length = measure_length
        self.time_resolution = time_resolution
        if self.time_resolution[1] > self.measure_length:
            print("Warning: Max duration exceeds measure length. Decreasing max duration to measure length.")
            self.time_resolution[1] = self.measure_length

        all_durations = [0.125, 0.25, 0.5, 1, 2, 3, 4, 8]
        self.possible_durations = [d for d in all_durations if self.time_resolution[0] <= d <= self.time_resolution[1]]
        # Set the weights and targets for each metric are set to default values if not provided
        if weights is None:
            self.weights = {
            'gini': (1.0, 1.0, 0.0),
            'balance': (1.0, 1.0, 0.0),
            'motif': (10.0, 1.0, 0.0),
            'dissonance': (1.0, 0.0, 0.0),
            'rhythmic': (0, 10.0, 0),
            'rest': (1.0, 0.0, 0.0)
        }
        else:
            self.weights = weights
        if targets is None:
            self.targets = {
            'gini': (0.05, 0.5, 0.0),  # Example targets; adjust as needed
            'balance': (0.1, 0.1, 0.0),
            'motif': (1.0, 1.0, 0.0),
            'dissonance': (0.0, 0.0, 0.0),
            'rhythmic': (0.0, 1.0, 0.0),  # Assuming rhythmic applies to duration
            'rest': (0.0, 0.0, 0.0) # rests are None pitches
        }
        else:
            self.targets = targets
        
        if seed is not None:
            random.seed(seed)

        self.best_individuals = []  # List to store the best individual of each generation
        self.best_scores = []
    
    def calculate_fitness_components(self, phrase):
        """
        Calculate the fitness components based on the Index class for a given phrase.

        Args:
            phrase (list): The musical phrase.

        Returns:
            dict: A dictionary containing the fitness components ('gini', 'balance', 'motif', 'dissonance', 'rhythmic').
        """
        fitness_components = {}
        # Split the phrase into its components
        pitches, durations, offsets = zip(*phrase)

        # Gini coefficient for pitch and duration
        fitness_components['gini_pitch'] = Index(pitches).gini()
        fitness_components['gini_duration'] = Index(durations).gini()

        # Balance for pitch and duration
        fitness_components['balance_pitch'] = Index(pitches).balance()
        fitness_components['balance_duration'] = Index(durations).balance()

        # Motif for pitch and duration (and potentially offsets if applicable)
        fitness_components['motif_pitch'] = Index(pitches).motif()
        fitness_components['motif_duration'] = Index(durations).motif()  # Add this if you're analyzing motifs in durations too

        # Dissonance, only applicable to pitches
        if self.scale is not None:
            fitness_components['dissonance_pitch'] = Index(pitches).dissonance(self.scale)

        # Rhythmic fit based on durations
        # Note: Here, I am assuming you have a method in Index class for rhythm analysis. Adjust if necessary.
        fitness_components['rhythmic'] = Index(durations).rhythmic(self.measure_length)  # Adjust method as needed

        # gini and balance for offsets?
        #fitness_components['gini_offset'] = Index(offsets).gini()  # Add if relevant
        #fitness_components['balance_offset'] = Index(offsets).balance()  # Add if relevant

        # Calculate the proportion of rest notes (notes where pitch is None)
        rest_notes = [note for note in phrase if note[0] is None]
        total_notes = len(phrase)
        rest_proportion = len(rest_notes) / total_notes if total_notes > 0 else 0
        fitness_components['rest'] = rest_proportion

        return fitness_components


    def fitness(self, phrase):
        """
        Calculate the fitness of a phrase based on how closely it meets the desired musical metrics targets.

        Args:
            phrase (list): Musical phrase.

        Returns:
            float: Fitness score of the phrase.
        """
        fitness_score = 0.0
        components = self.calculate_fitness_components(phrase)

        # Iterate through each metric and calculate the contribution to fitness
        for metric, (pitch_target, duration_target, offset_target) in self.targets.items():
            for i, component in enumerate(['pitch', 'duration', 'offset']):
                actual_value = components.get(f"{metric}_{component}", 0)
                target_value = (pitch_target, duration_target, offset_target)[i]
                weight = self.weights[metric][i]

                # Calculate the similarity from the target (only if there is a target)
                if target_value is not None and weight > 0:  # Ensure there's a target and weight for this component
                    similarity = 1 - abs(actual_value - target_value) / (target_value if target_value != 0 else 1)
                    fitness_score += max(0, similarity) * weight  # Apply weight and ensure non-negative score

        return fitness_score

    def mutate(self, phrase, rate=None, rest_rate=0.02):
        """
        Mutate a musical phrase while respecting musical structures and boundaries.
        """
        if rate is None:
            rate = self.mutation_rate

        new_phrase = []
        total_offset = 0  # Track total offset for alignment

        for note in phrase:
            pitch, duration, offset = note

            # Mutate pitch, duration, and rest based on mutation probabilities
            new_pitch = self.mutation_probabilities['pitch']() if random.random() < rate else pitch
            new_duration = self.mutation_probabilities['duration']() if random.random() < rate else duration
            new_rest = self.mutation_probabilities['rest']()

            # Align offsets based on sequential order, ensuring notes/rests align with previous ones
            new_offset = total_offset
            total_offset += new_duration  # Increment total offset by new duration

            # Append the new note or rest to the phrase
            new_phrase.append((new_pitch if new_rest else None, new_duration, new_offset))

        return new_phrase

    def select(self, k=25):
        """
        Select top-k phrases based on fitness.

        Args:
            k (int): Number of top phrases to select.

        Returns:
            list: Top k musical phrases.
        """
        # Evaluate the fitness of each phrase in the population
        fitness_scores = [(phrase, self.fitness(phrase)) for phrase in self.population]
        # Sort the phrases by their fitness scores in descending order
        sorted_phrases = sorted(fitness_scores, key=lambda x: x[1], reverse=True)
        # Select the top-k phrases
        selected_phrases = [phrase for phrase, score in sorted_phrases[:k]]
        return selected_phrases

    def crossover(self, parent1, parent2):
        """
        Combine two phrases to create a new phrase.

        Args:
            parent1, parent2 (list): Parent phrases.

        Returns:
            list: New musical phrase generated from parents.
        """
        # Cut points for the crossover
        cut1, cut2 = sorted(random.sample(range(len(parent1)), 2))
        # Create new phrase by combining parts of the parent phrases
        new_phrase = parent1[:cut1] + parent2[cut1:cut2] + parent1[cut2:]
        return new_phrase

    def evolve(self, k=25, rest_rate=0.02):
        """
        Evolve the population of phrases through selection, crossover, and mutation.

        Args:
            k (int): Number of phrases to select for the next generation.
        """
        # Selection based on fitness
        selected_phrases = self.select(k)

        # Store the best individual and its fitness score
        self.best_individuals.append(selected_phrases[0])
        self.best_scores.append(self.fitness(selected_phrases[0]))

        # Crossover and mutation to create a new generation
        new_population = []
        while len(new_population) < self.population_size:
            parent1, parent2 = random.sample(selected_phrases, 2)
            parent1 = copy.deepcopy(parent1)
            parent2 = copy.deepcopy(parent2)
            child = self.crossover(parent1, parent2)
            mutated_child = self.mutate(child, rest_rate=rest_rate)
            new_population.append(mutated_child)

        self.population = new_population


import os
import itertools
import warnings
import json
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.axes
from . import utils

try:
    from numba import jit
    NUMBA_AVAILABLE = True
except ImportError:
    NUMBA_AVAILABLE = False
    def jit(*args, **kwargs):
        def decorator(func):
            return func
        return decorator

def optional_jit(*args, **kwargs):
    """
    Applies a jit decorator only if numba is installed
    """
    def decorator(func):
        if NUMBA_AVAILABLE:
            return jit(*args, **kwargs)(func)
        return func
    return decorator


# Cellular automata
# -----------------
class CellularAutomata:
    """
    A class for simulating one-dimensional cellular automata based on a specific rule set.

    Args:
        rule_number (int or str): Rule number for the cellular automaton, must be between 0 and 255.
        width (int): Number of cells in the automaton's width.
        initial_state (list of int, optional): Initial state of the automaton. Defaults to all zeros with a central one.

    Attributes:
        width (int): The cellular automaton's width.
        rule_number (str): The rule number, zero-padded to three digits.
        initial_state (list of int): The initial state of the automaton.
        state (list of int): The current state of the automaton.
        rules (list of tuple): The rules loaded from a JSON file.
    """
    
    def __init__(self, rule_number, width, initial_state=None):
        """
        Initializes the CellularAutomata class with a rule number, width, and optionally an initial state.

        Parameters:
            rule_number (int or str): The rule number for the cellular automaton, must be between 0 and 255.
            width (int): The number of cells in one row of the automaton.
            initial_state (list, optional): Initial binary state of the automaton. Defaults to a list of zeros with a single one in the center.

        Raises:
            ValueError: If the rule number is not within the valid range (0 to 255).
        """
        self.width = width
        if int(rule_number) < 0 or int(rule_number) > 255:
            raise ValueError("Rule number must be an integer between 0 and 255.")
        if isinstance(rule_number, int):
            rule_number = str(rule_number).zfill(3)
        self.rule_number = rule_number
        self.rules = self.load_rules(rule_number)
        self.initial_state = [0] * width if initial_state is None else initial_state
        self.state = self.initial_state.copy()

    def load_rules(self, rule_number):
        """
        Loads the rules from a JSON file based on the rule number.

        Parameters:
            rule_number (str): The rule number as a string, zero-padded to three digits.

        Returns:
            list: A list of tuples representing the rules for the cellular automaton.
        """
        script_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(script_dir, 'data', 'ca1D_rules.json')

        with open(file_path, 'r') as file:
            rules_json = json.load(file)
        rule_conditions = rules_json[rule_number]
        return [tuple(condition) for condition in rule_conditions]

    def update_state(self):
        """
        Updates the state of the automaton based on its rules for one generation.
        """
        new_state = [0] * self.width
        for i in range(self.width):
            neighborhood = (self.state[(i - 1) % self.width],
                            self.state[i],
                            self.state[(i + 1) % self.width])
            new_state[i] = 1 if neighborhood in self.rules else 0
        self.state = new_state

    def validate_strips(self, strips):
        """
        Validates that the strips are correctly formatted and within the valid range.

        Parameters:
            strips (list of tuples): Each tuple should specify the start and end indices of a strip in the automaton.

        Raises:
            ValueError: If strips are not properly formatted or indices are out of range.
        """
        if isinstance(strips, tuple):
            strips = [strips]
        if not all(isinstance(strip, tuple) for strip in strips) or not all(len(strip) == 2 for strip in strips) or not all(isinstance(cell, int) for strip in strips for cell in strip):
            raise ValueError("Strips must be a tuple of two integers or a list of tuples with two integers each.")
        if not all(0 <= strip[0] < strip[1] <= self.width for strip in strips): 
            raise ValueError("Strip values must be within the width of the automata.")

    def validate_values(self, values, strips):
        """
        Validates that the values are provided as dictionaries, and there is a dictionary for each strip.

        Parameters:
            values (list or dict): List of dictionaries mapping indices to pitches or other data.
            strips (list of tuples): List of strip ranges to validate against.

        Raises:
            ValueError: If values are not provided as a list of dictionaries or their count doesn't match the strips.
        """
        if isinstance(values, dict):
            values = [values]
        if not isinstance(values, list) or not all(isinstance(val_dict, dict) for val_dict in values):
            raise ValueError("Values must be provided as a dictionnary or a list of dictionaries mapping indices to pitches or other data.")
        if len(values) != len(strips):
            raise ValueError("The number of value dictionaries must match the number of strips.")

    def generate_01(self, iterations, strips=None):
        """
        Generates a binary (0 or 1) evolution of the automaton over a specified number of iterations, optionally for specific strips.

        Parameters:
            iterations (int): Number of iterations to evolve the automaton.
            strips (list of tuples, optional): Specific sections of the automaton to evolve.

        Returns:
            list: A list representing the evolution of the automaton, either as a whole or just the specified strips.
        """
        if strips:
            self.validate_strips(strips)
        self.state = self.initial_state.copy()
        evolution = [self.state.copy()]
        for _ in range(iterations - 1):
            self.update_state()
            evolution.append(self.state.copy())
        
        if strips:
            strip_evolutions = []
            for strip in strips:
                strip_evolution = [row[strip[0]:strip[1]] for row in evolution]
                strip_evolutions.append(strip_evolution)
            return strip_evolutions
        
        return evolution
    
    def generate(self, iterations, strips, values):
        """
        Generates the evolution of the automaton, mapping binary states (0 or 1) to specific values based on provided mappings.

        Parameters:
            iterations (int): Number of iterations to evolve the automaton.
            strips (list of tuples): Sections of the automaton for which to generate values.
            values (list of dicts): Mappings from indices in each strip to specific values.

        Returns:
            list: A list of evolutions for each strip, with binary states mapped to specified values.
        """

        # Validate strips and value mapping
        self.validate_strips(strips)
        self.validate_values(values, strips)
        if isinstance(values, dict):
            values = [values]

        # Generate the binary evolution for specified strips
        evolution_01 = self.generate_01(iterations, strips)
        values_evolution = []
        for strip_evolution, value_dict in zip(evolution_01, values):  # Process each strip with its corresponding value map
            strip_values = []
            for row in strip_evolution:
                row_values = []
                for i, cell in enumerate(row):
                    if cell == 1 and i in value_dict:
                        row_values.append(value_dict[i])
                if row_values:
                    if len(row_values) == 1:
                        strip_values.append(row_values[0])  # Single value
                    else:
                        strip_values.append(row_values)  # Multiple values forming a chord
                else:
                    strip_values.append(None)
            values_evolution.append(strip_values)
        if len(values_evolution) == 1:
            return values_evolution[0]
        else:
            return values_evolution

    def plot(self, iterations, ax=None, strips=None, extract_strip=False, title=None, show_axis=True):
        """
        Plots the evolution of the cellular automaton.

        Parameters:
            iterations (int): Number of generations to simulate.
            ax (matplotlib.axes.Axes, optional): The matplotlib axis to plot on. If None, a new figure is created.
            strips (list of tuples, optional): Ranges to highlight or exclusively plot.
            extract_strip (bool): If True, only the specified strips are plotted each in separate subplots.
            title (str, optional): Title for the plot. Default is based on the rule number.
            show_axis (bool): Whether to show axis labels and grid.

        Returns:
            matplotlib.axes.Axes: The axis with the plot.
        """

        # Handle based on the extraction flag
        if not extract_strip:
            if title is None:
                title = f"Cellular Automata Evolution for Rule {self.rule_number}"
            # Display the complete evolution
            if ax is None:
                fig, ax = plt.subplots(figsize=(12, 8))
            evolution = np.array(self.generate_01(iterations)).T
            im = ax.imshow(evolution, cmap='binary', aspect='equal')
            ax.invert_yaxis()
            ax.set_title(title)
            ax.set_ylabel("Cell Position")
            ax.set_xlabel("Generation")

            # Add strips if provided
            if strips:
                label_offset = 3
                for index, strip in enumerate(strips):
                    rect = patches.Rectangle(
                        (0, strip[0]),
                        iterations, strip[1] - strip[0], linewidth=1, edgecolor='none', facecolor='#88888880'
                    )
                    ax.add_patch(rect)
                    ax.text(
                        1, strip[0] + label_offset, f'Strip {index + 1}',
                        color='white', fontsize=10, verticalalignment='top',
                        bbox=dict(facecolor='#333333', edgecolor='none', pad=2)
                    )
        else:
            # Handle multiple strips, each in its own subplot
            if not strips or len(strips) == 1:
                fig, axs = plt.subplots(figsize=(12, 8))  # Use a new axis for a single strip
            else:
                fig, axs = plt.subplots(len(strips), 1, figsize=(12, 2 * len(strips)**(0.5) ))
                axs = axs.flatten()

            if isinstance(axs, matplotlib.axes.Axes):
                axs = [axs]

            strip_evolutions = self.generate_01(iterations, strips)
            for i, strip_data in enumerate(strip_evolutions):
                axs[i].imshow(np.array(strip_data).T, cmap='binary', aspect='equal')
                axs[i].invert_yaxis()
                if title is None:
                    strip_title = f"Strip {i + 1} Evolution" 
                else:
                    strip_title = title[i]
                axs[i].set_title(strip_title)
                axs[i].set_ylabel("Cell Position")
                axs[i].set_xlabel("Generation")

        # Optionally turn off axis
        if not show_axis:
            ax.axis('off')

        # If this function created the figure and it's handling a subplot setup, adjust the layout
        if extract_strip:
            plt.tight_layout()
            plt.subplots_adjust(hspace=0.5)  # Adjust horizontal space if needed

        return ax
    

# Mandelbrot fractal
# ------------------
@optional_jit(nopython=True)
def generate_mandelbrot_jit(x_range, y_range, dimensions, max_iter):
    x_lin = np.linspace(x_range[0], x_range[1], dimensions[0])
    y_lin = np.linspace(y_range[0], y_range[1], dimensions[1])
    output = np.zeros(dimensions, dtype=np.int32)  # Initialize the output array

    for i in range(dimensions[0]):
        for j in range(dimensions[1]):
            x = x_lin[i]
            y = y_lin[j]
            C = complex(x, y)  # Ensure using complex number
            Z = 0 + 0j  # Initialize Z as a complex zero
            count = 0  # Initialize escape time count

            while abs(Z) < 2 and count < max_iter:
                Z = Z**2 + C
                count += 1
            
            output[i, j] = count  # Set the output based on the escape time
    
    return output

class Mandelbrot:
    def __init__(self, scale=None, start_note_index=0, dimensions=(800, 800), max_iter=1000, x_range=(-2.0, 1.0), y_range=(-1.5, 1.5)):
        if isinstance(dimensions, int):
            dimensions = (dimensions, dimensions)
        if not isinstance(dimensions, tuple) or not isinstance(max_iter, int):
            raise ValueError("Dimensions must be a tuple and max_iter must be an integer.")
        if dimensions[0] <= 0 or dimensions[1] <= 0 or max_iter <= 0:
            raise ValueError("Dimensions and max_iter must be positive.")
        
        self.scale = scale
        self.start_note_index = start_note_index
        self.dimensions = dimensions
        self.max_iter = max_iter
        self.x_range = x_range
        self.y_range = y_range

    def generate_mandelbrot(self):
        return generate_mandelbrot_jit(self.x_range, self.y_range, self.dimensions, self.max_iter)
    
    def generate(self, method='horizontal', line_index=0):
        data = self.generate_mandelbrot()
        if method == 'horizontal':
            return data[line_index, :]
        elif method == 'vertical':
            return data[:, line_index]
        elif method == 'diagonal-increasing':
            return np.diagonal(data)
        elif method == 'diagonal-decreasing':
            return np.diagonal(np.flipud(data))
        elif method == 'random':
            flat_data = data.flatten()
            return np.random.choice(flat_data, size=100, replace=False)

    def plot(self, ax=None, figsize=(10, 10), zoom_rect=None, show_numbers=False):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        fractal = self.generate_mandelbrot()
        extent = (self.x_range[0], self.x_range[1], self.y_range[0], self.y_range[1])
        if zoom_rect and show_numbers:
            warnings.warn("Both zoom rectangle and showing numbers are enabled. Numbers are hidden.", UserWarning)
        im = ax.imshow(
            fractal.T,
            extent=extent,
            cmap='viridis', aspect='auto', origin='lower'
        )
        ax.set_title('Mandelbrot Set')
        ax.set_xlabel('Real')
        ax.set_ylabel('Imaginary')

        if show_numbers:
            for i in range(fractal.shape[0]):
                for j in range(fractal.shape[1]):
                    ax.text(j, i, fractal[i, j], ha="center", va="center", color="w")

        if zoom_rect:
            rect = patches.Rectangle(
                (zoom_rect[0][0], zoom_rect[1][0]),
                zoom_rect[0][1] - zoom_rect[0][0], zoom_rect[1][1] - zoom_rect[1][0],
                linewidth=1, edgecolor='white', facecolor='none'
            )
            ax.add_patch(rect)
        return ax

# Logistic map
# ------------
@optional_jit(nopython=True)
def logistic_map(growth_rate, pop, iterations):
    """Compute logistic map iteratively for a given rate over many iterations."""
    final_pop = np.empty(iterations)
    for i in range(iterations):
        pop = growth_rate * pop * (1 - pop)
        final_pop[i] = pop
    return final_pop

@optional_jit(nopython=True)
def compute_logistic(rate_values, iterations, last_n):
    """Compute the logistic map for a range of r values, collecting the last `last_n` iterations."""
    num_rate = len(rate_values)
    plot_pop = np.empty(num_rate * last_n)
    plot_rate = np.empty(num_rate * last_n)
    for idx, r in enumerate(rate_values):
        xs = logistic_map(r, 0.5, iterations + last_n)  # Drop initial values to skip transient
        plot_pop[idx*last_n:(idx+1)*last_n] = xs[-last_n:]  # Take only the last `last_n` iterations
        plot_rate[idx*last_n:(idx+1)*last_n] = r
    return plot_rate, plot_pop

class LogisticMap:
    def __init__(self, rates, iterations=1000, last_n=100):
        self.rates = rates
        self.iterations = iterations
        self.last_n = last_n
    
    def generate(self):
        rate, pop = compute_logistic(self.rates, self.iterations, self.last_n)
        return rate, pop
    
    def plot(self, ax=None, figsize=(10, 6)):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        rate, pop = self.generate()
        ax.plot(rate, pop, ',k', alpha=0.5)
        ax.set_xlabel('rate')
        ax.set_ylabel('population')
        ax.set_xlim(self.rates[0], self.rates[-1])
        ax.set_ylim(0, 1)
        return ax


from . import analysis
from . import conversion
from . import fractal
from . import genetic
from . import harmony
from . import loop
from . import minimalism
from . import player
from . import rhythm
from . import score
from . import utils
from . import walk

#from importlib.util import find_spec
#import warnings
#tensorflow_spec = find_spec('tensorflow')
#if tensorflow_spec is not None:
#    from . import djai
#else:
#    warnings.warn("Tensorflow is not installed. Djai cannot be imported.")

__version__ =  "0.2.1"


import random
import numpy as np
from . import utils
import itertools

def isorhythm(pitches, durations):
    """
    Merges durations and pitches until both ends coincide, then sets offsets according to successive durations.

    Args:
        pitches (list): The first list.
        durations (list): The second list.

    Returns:
        list: A list of notes.
    """
    lcm = np.lcm(len(pitches), len(durations))

    p_repeated = (pitches * (lcm // len(pitches)))[:lcm]
    d_repeated = (durations * (lcm // len(durations)))[:lcm]
    o = [1] * lcm
    notes = list(zip(p_repeated, d_repeated, o))
    notes = utils.set_offsets_according_to_durations(notes)

    return notes

def beatcycle(pitches, durations):
    """
    Pitches are mapped to durations in a cyclical manner, then offsets are set according to successive durations.

    Args:
        pitches (list): The first list.
        durations (list): The second list.

    Returns:
        list: A list of notes.
    """
    durations_cycle = itertools.cycle(durations)
    notes = []
    current_offset = 0
    for p in pitches:
        d = next(durations_cycle)
        notes.append((p, d, current_offset))
        current_offset += d
    return notes

class Rhythm:
    """
    A class used to represent a Rhythm.

    Attributes:
        measure_length (int): the length of the measure
        durations (list): the durations of the notes
    """

    def __init__(self, measure_length, durations):
        """
        Constructs all the necessary attributes for the Rhythm object.

        Args:
            measure_length (int): the length of the measure
            durations (list): the durations of the notes
        """
        self.measure_length = measure_length
        self.durations = durations

    def random(self, seed=None, rest_probability=0, max_iter=100):
        """
        Generate a random rhythm as a list of (duration, offset) tuples.

        Args:
            duration (list): List of possible durations.
            measure_length (float): Total length of the measure.
            rest_probability (float): Probability of a rest (i.e., removing a tuple).
            max_iter (int): Maximum number of iterations to generate the rhythm.

        Returns:
            list: List of (duration, offset) tuples representing the rhythm.
        """
        random.seed(seed)
        rhythm = []
        total_length = 0.0
        n_iter = 0
        while total_length < self.measure_length:
            if n_iter >= max_iter:
                print('Max iterations reached. The sum of the durations is not equal to the measure length.')
                break # Avoid infinite loops
            d = random.choice(self.durations)
            if total_length + d > self.measure_length:
                continue
            if random.random() < rest_probability:
                continue
            rhythm.append((d, total_length))
            total_length += d
            n_iter += 1
        return rhythm
    
    def darwin(self, seed=None, population_size=10, max_generations=50, mutation_rate=0.1):
        """
        Executes the Darwinian evolution algorithm to generate the best rhythm.

        Args:
            seed (int): The random seed for reproducibility.
            population_size (int): The number of rhythms in each generation.
            max_generations (int): The maximum number of generations to evolve.
            mutation_rate (float): The probability of mutating a given rhythm.

        Returns:
            list: The best rhythm found after the last generation, sorted by ascending offset.
        """

        ga = GeneticRhythm(seed, population_size, self.measure_length, max_generations, mutation_rate, self.durations)
        best_rhythm = ga.generate()
        return best_rhythm


class GeneticRhythm:
    def __init__(self, seed, population_size, measure_length, max_generations, mutation_rate, durations):
        """
        Initializes a Rhythm Genetic Algorithm instance.
        
        Args:
            population_size (int): The number of rhythms in each generation.
            measure_length (int): The total length of the rhythm to be generated.
            max_generations (int): The maximum number of generations to evolve.
            mutation_rate (float): The probability of mutating a given rhythm.
            durations (list): List of possible note durations.
        """
        random.seed(seed)
        self.population_size = population_size
        self.measure_length = measure_length
        self.max_generations = max_generations
        self.mutation_rate = mutation_rate
        self.durations = durations
        self.population = self.initialize_population()

    def initialize_population(self):
        """Initializes a population of random rhythms."""
        population = []
        for _ in range(self.population_size):
            rhythm = self.create_random_rhythm()
            population.append(rhythm)
        return population

    def create_random_rhythm(self):
        """
        Creates a random rhythm ensuring it respects the measure length and has no overlapping notes.
        
        Returns:
            list: A list of (duration, offset) tuples representing the rhythm.
        """
        rhythm = []
        total_length = 0
        while total_length < self.measure_length:
            remaining = self.measure_length - total_length
            note_length = random.choice(self.durations)
            if note_length <= remaining:
                rhythm.append((note_length, total_length))
                total_length += note_length
        return rhythm

    def evaluate_fitness(self, rhythm):
        """
        Evaluates the fitness of a rhythm based on how close it is to the total measure length.
        
        Args:
            rhythm (list): The rhythm to evaluate, represented as a list of (duration, offset) tuples.
        
        Returns:
            int: The fitness score of the rhythm.
        """
        total_length = sum(note[0] for note in rhythm)  # Use note[0] for duration
        return abs(self.measure_length - total_length)

    def select_parents(self):
        """
        Selects two parents for reproduction using a simple random selection approach.
        
        Returns:
            tuple: Two selected parent rhythms for crossover.
        """
        parent1 = random.choice(self.population)
        parent2 = random.choice(self.population)
        return parent1 if self.evaluate_fitness(parent1) < self.evaluate_fitness(parent2) else parent2

    def crossover(self, parent1, parent2):
        """
        Performs crossover between two parent rhythms to produce a new child rhythm.
        
        Args:
            parent1 (list): The first parent rhythm.
            parent2 (list): The second parent rhythm.
        
        Returns:
            list: The new child rhythm generated from the parents.
        """
        crossover_point = random.randint(1, len(parent1) - 1)
        child = parent1[:crossover_point] + parent2[crossover_point:]
        return self.ensure_measure_length(child)

    def ensure_measure_length(self, rhythm):
        """
        Ensures that the rhythm respects the measure length, adjusting if necessary.
        
        Args:
            rhythm (list): The rhythm to check and adjust.
        
        Returns:
            list: The adjusted rhythm.
        """
        total_length = sum(note[0] for note in rhythm)  # Changed to note[0] for duration since we're working with (duration, offset) now
        if total_length > self.measure_length:
            rhythm.pop()  # Remove the last note if the total duration exceeds the measure length
        return rhythm

    def mutate(self, rhythm):
        """
        Performs a mutation on a rhythm with a certain probability, ensuring no note overlap.
        
        Args:
            rhythm (list): The rhythm to mutate.
        
        Returns:
            list: The mutated rhythm.
        """
        if random.random() < self.mutation_rate:
            index = random.randint(0, len(rhythm) - 2)  # Avoid mutating the last note for simplicity
            duration, offset = rhythm[index]
            next_offset = self.measure_length if index == len(rhythm) - 1 else rhythm[index + 1][1]
            max_new_duration = next_offset - offset
            new_durations = [d for d in self.durations if d <= max_new_duration]
            if new_durations:
                new_duration = random.choice(new_durations)
                rhythm[index] = (new_duration, offset)
        return rhythm

    def generate(self):
        """
        Executes the genetic algorithm, evolving the rhythms over generations.
        
        Returns:
            list: The best rhythm found after the last generation, sorted by ascending offset.
        """
        for generation in range(self.max_generations):
            new_population = []
            for _ in range(self.population_size):
                parent1 = self.select_parents()
                parent2 = self.select_parents()
                child = self.crossover(parent1, parent2)
                child = self.mutate(child)
                child_sorted = sorted(child, key=lambda x: x[1])  # x[1] est l'offset dans le tuple (duration, offset)
                new_population.append(child_sorted)
            self.population = new_population

        best_rhythm = min(self.population, key=self.evaluate_fitness)
        best_rhythm_sorted = sorted(best_rhythm, key=lambda x: x[1])  # Trier par offset ascendant
        return best_rhythm_sorted





def round_to_list(value, scale):
    """
    Rounds the given value to the nearest value in the scale list.

    Args:
        value (float): The value to be rounded.
        scale (list): A list of values to round to.

    Returns:
        float: The value from the scale list that is closest to the given value.
    """
    return min(scale, key=lambda x: abs(x - value))

def get_octave(midi_note):
    return midi_note // 12 - 1

def get_sharp(string):
    dict_flat = {'D-': 'C#', 'E-': 'D#', 'G-': 'F#', 'A-': 'G#', 'B-': 'A#'}
    if string in dict_flat.keys():
        string = dict_flat[string]
    return string

def get_degree_from_pitch(pitch, scale_list, tonic_pitch):

    if isinstance(pitch, str):
        pitch = cde_to_midi(pitch)
    if isinstance(tonic_pitch, str):
        tonic_pitch = cde_to_midi(tonic_pitch) 

    tonic_index = scale_list.index(tonic_pitch)

    # If the pitch is in the mode
    if pitch in scale_list:
        # Find its index and compute the degree
        pitch_index = scale_list.index(pitch)
        degree = pitch_index - tonic_index
    else:
        # If the pitch is not in the mode, find the two pitches it falls between
        upper_pitch = round_to_list(pitch, scale_list)
        upper_index = scale_list.index(upper_pitch)
        lower_index = upper_index - 1 if upper_index > 0 else upper_index
        lower_pitch = scale_list[lower_index]

        # Compute the degree as the weighted average of the degrees of the two pitches
        distance_to_upper = upper_pitch - pitch
        distance_to_lower = pitch - lower_pitch
        upper_weight = 1 - distance_to_upper / (distance_to_upper + distance_to_lower)
        lower_weight = 1 - distance_to_lower / (distance_to_upper + distance_to_lower)
        upper_degree = upper_index - tonic_index
        lower_degree = lower_index - tonic_index
        degree = upper_degree * upper_weight + lower_degree * lower_weight

    return degree

def get_pitch_from_degree(degree, scale_list, tonic_pitch):
    tonic_index = scale_list.index(tonic_pitch)
    pitch_index = round(tonic_index + degree)  # round to nearest integer

    # If the degree is within the scale
    if 0 <= pitch_index < len(scale_list):
        pitch = scale_list[pitch_index]
    else:
        # If the degree is not within the scale, find the two pitches it falls between
        lower_index = max(0, min(pitch_index, len(scale_list) - 1))
        upper_index = min(len(scale_list) - 1, max(pitch_index, 0))
        lower_pitch = scale_list[lower_index]
        upper_pitch = scale_list[upper_index]

        # Compute the pitch as the weighted average of the two pitches
        distance_to_upper = upper_index - pitch_index
        distance_to_lower = pitch_index - lower_index
        if distance_to_upper + distance_to_lower == 0:
            upper_weight = lower_weight = 0.5
        else:
            upper_weight = 1 - distance_to_upper / (distance_to_upper + distance_to_lower)
            lower_weight = 1 - distance_to_lower / (distance_to_upper + distance_to_lower)
        pitch = upper_pitch * upper_weight + lower_pitch * lower_weight

    return pitch

def set_offsets_according_to_durations(notes):
    """
    Adjusts the offsets of the notes based on their durations.

    Args:
        notes (list): A list of tuples, where each tuple contains a note (pitch),
                      a duration (quarterLength), and an offset.

    Returns:
        list: The list of notes with adjusted offsets.
    """
    if len(notes[0]) == 2:
        notes = [(note[0], note[1], 0) for note in notes]
    adjusted_notes = []
    current_offset = 0

    for pitch, duration, _ in notes:
        adjusted_notes.append((pitch, duration, current_offset))
        current_offset += duration

    return adjusted_notes

def fill_gaps_with_rests(notes, parent_offset=0.0):
    """
    Analyze a sorted list of notes (each note is a (pitch, duration, offset) tuple)
    and insert rests (None, duration, offset) to fill gaps between notes. Notes are
    sorted by offset before processing to ensure accurate gap detection and filling.

    Args:
        notes (list): The list of notes to be processed, not necessarily sorted.
        parent_offset (float): The offset to consider from the parent sequence, used in recursion.

    Returns:
        list: The modified list with gaps filled with rests, ensuring continuity.
    """
    # Sort notes by offset to ensure correct processing order
    notes_sorted = sorted(notes, key=lambda x: x[2])

    last_offset = 0.0  # Keep track of the offset after the last note or rest
    filled_notes = []

    for note in notes_sorted:
        pitch, duration, offset = note
        current_offset = parent_offset + offset
        if current_offset > last_offset:
            # There is a gap that needs to be filled with a rest
            gap_duration = current_offset - last_offset
            rest_to_insert = (None, gap_duration, last_offset - parent_offset)
            filled_notes.append(rest_to_insert)  # Insert the rest to fill the gap

        filled_notes.append(note)
        last_offset = max(last_offset, current_offset + duration)  # Update last offset for the next iteration

    return filled_notes


def adjust_note_durations_to_prevent_overlaps(notes):
    """
    Adjust the durations of notes in a list (each note is a (pitch, duration, offset) tuple)
    to prevent overlaps, while keeping their offsets intact.

    Args:
        notes (list): The list of notes to be adjusted.

    Returns:
        list: The modified list with adjusted note durations.
    """
    # Ensure the list is sorted by offset
    notes.sort(key=lambda note: note[2])

    for i in range(len(notes) - 1):  # Loop through all notes except the last one
        current_note = notes[i]
        next_note = notes[i + 1]
        # Calculate the current end of the note
        current_note_end = current_note[2] + current_note[1]

        # If the current note ends after the next note starts, adjust its duration
        if current_note_end > next_note[2]:
            # Adjust duration to avoid overlap
            new_duration = next_note[2] - current_note[2]
            notes[i] = (current_note[0], new_duration, current_note[2])

    return notes


def repair_notes(s: list) -> list:
    """
    Apply the fill_gaps_with_rests and adjust_note_durations_to_prevent_overlaps functions to a stream.

    Args:
        s (stream.Stream): The music21 stream to be processed.

    Returns:
        stream.Stream: The modified stream with gaps filled and note durations adjusted.
    """
    return adjust_note_durations_to_prevent_overlaps(fill_gaps_with_rests(s))


def cde_to_midi(pitch):
    # Mapping of note names to MIDI numbers with sharps
    pitches = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
    flat_to_sharp = {
        'Db': 'C#', 'Eb': 'D#', 'Gb': 'F#', 'Ab': 'G#', 'Bb': 'A#',
        'Cb': 'B',  # Special handling for 'Cb', converting it directly to 'B'
    }

    octave = 4  # Default octave if not specified in the pitch string

    # Check and convert flat notes to sharp notes
    if 'b' in pitch:
        note = pitch[:-1]  # Exclude the octave number if present
        if note in flat_to_sharp:
            pitch = flat_to_sharp[note] + pitch[-1]  # Append the octave number back if it was present

    # Extract the note (with sharp) and octave from the pitch
    if len(pitch) > 2 or pitch[1].isdigit():
        note, octave = pitch[:-1], int(pitch[-1])
    else:
        note = pitch[0]

    midi = 12 * (octave + 1) + pitches.index(note)
    return midi


def midi_to_cde(midi):
    # Mapping of MIDI numbers to note names
    pitches = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
    octave = midi // 12 - 1
    key = midi % 12
    pitch = pitches[key] + str(octave)
    return pitch


def no_overlap(notes, adjust='offsets'):
    """
    Adjusts the offsets of the notes to prevent overlap.

    Args:
        notes (list): A list of tuples, where each tuple contains a note (pitch),
                      a duration (quarterLength), and an offset.

    Returns:
        list: The list of notes with adjusted offsets or durations.
    """
    adjusted_notes = []
    current_offset = 0

    for pitch, duration, _ in notes:
        adjusted_notes.append((pitch, duration, current_offset))
        current_offset += duration

    return adjusted_notes


def check_input(input_list):
    """
    Checks if the input is a list of simple elements or a list of tuples.

    Args:
        input_list (list): The input to check.

    Returns:
        str: 'list' if the input is a list of simple elements,
             'list of tuples' if the input is a list of tuples,
             'unknown' otherwise.
    """
    if all(isinstance(i, tuple) for i in input_list):
        return 'list of tuples'
    elif all(not isinstance(i, tuple) for i in input_list):
        return 'list'
    else:
        return 'unknown'


def scale_list(numbers, to_min, to_max, min_numbers=None, max_numbers=None):
    """
    Scale a list of numbers so that its range is between min_value and max_value.

    Args:
        numbers (list): List of numbers to scale.
        min_value (float): Minimum value of the scaled list.
        max_value (float): Maximum value of the scaled list.

    Returns:
        list: Scaled list of numbers.
    """
    if min_numbers is None:
        min_numbers = min(numbers)
    if max_numbers is None:
        max_numbers = max(numbers)
    if min_numbers == max_numbers:
        return [(min_numbers + max_numbers) / 2] * len(numbers)
    else:
        return [(num - min_numbers) * (to_max - to_min) / (max_numbers - min_numbers) + to_min for num in numbers]


def offset_track(track, by):
    """
    Offset the notes in a list by a given amount.

    Args:
        track (list): List of notes to offset.
        by (float): Amount to offset the notes.

    Returns:
        list: List of notes with adjusted offsets.
    """
    return [(pitch, duration, offset + by) for pitch, duration, offset in track]


def quantize_notes(notes, measure_length, time_resolution):
    """
    Quantize the durations and offsets of notes in musical phrases.

    Args:
        notes (list): List of musical phrases, where each phrase is a list of tuples (pitch, duration, offset).
        measure_length (float): The total duration of a measure, typically in quarter notes.
        time_resolution (float): The smallest time unit for quantization, typically in quarter notes.

    Returns:
        list: The quantized musical phrases.
    """
    quantized_notes = []
    for note in notes:
        pitch, duration, offset = note
        quantized_offset = round(offset / time_resolution) * time_resolution
        measure_end = ((quantized_offset // measure_length) + 1) * measure_length
        quantized_duration = round(duration / time_resolution) * time_resolution
        quantized_duration = min(quantized_duration, measure_end - quantized_offset)
        quantized_notes.append((pitch, quantized_duration, quantized_offset))
    quantized_notes = [note for note in quantized_notes if note[1] > 0]  # remove notes with zero duration
    return quantized_notes


def find_closest_pitch_at_measure_start(notes, measure_length):
    """
    Finds the closest pitch at the beginning of each measure.

    Args:
        notes (list of tuples): A list of tuples where each tuple is (pitch, duration, offset).
        measure_length (float): The length of a measure.

    Returns:
        list: A list of pitches, each representing the closest pitch at the start of a measure.
    """
    # Filter out notes with None offset or pitch
    notes = [note for note in notes if note[2] is not None and note[0] is not None]

    # Sort the notes by offset to ensure they are in order
    notes_sorted_by_offset = sorted(notes, key=lambda x: x[2])

    # Find the maximum offset to determine how many measures we have
    max_offset = max(notes_sorted_by_offset, key=lambda x: x[2])[2]
    num_measures = int(max_offset // measure_length) + 1

    closest_pitches = []

    for measure_num in range(num_measures):
        measure_start = measure_num * measure_length
        closest_pitch = None
        closest_distance = float('inf')

        for pitch, duration, offset in notes_sorted_by_offset:
            # Calculate the distance from the start of the measure to the note's offset
            distance = measure_start - offset

            # If the note starts before the measure and is closer than any note we've looked at before
            if distance >= 0 and distance < closest_distance:
                closest_distance = distance
                closest_pitch = pitch

            # If we've passed the current measure start, we can break out of the loop
            if offset > measure_start:
                break

        if closest_pitch is not None:
            closest_pitches.append(closest_pitch)

    return closest_pitches

def tune(pitch, scale):
    """
    Adjust the pitch of a note to the nearest pitch within the given scale.

    Args:
        pitch (int): a MIDI pitch number to tune.
        scale (list): A list of pitches

    Returns:
        pitch: A tuned MIDI pitch number.
    """
    return min(scale, key=lambda x: abs(x - pitch))

def ql_to_seconds(ql, bpm):
    """
    Convert a duration in quarter-length units to seconds.

    Args:
        ql (float): Duration in quarter-length units.
        bpm (float): Beats per minute.

    Returns:
        float: Duration in seconds.
    """
    return 60 / bpm * ql

def fibonacci(a = 0, b = 1, base = 0, scale = 1):
    """
    Generate a Fibonacci iterator.

    This function generates a Fibonacci iterator that yields the next Fibonacci number in the sequence.

    Args:
        a (int): The first number in the Fibonacci sequence (default is 0).
        b (int): The second number in the Fibonacci sequence (default is 1).
        base (int): The base value to be added to each Fibonacci number (default is 0).
        scale (int): The scale factor to be multiplied with each Fibonacci number (default is 1).

    Yields:
        int: The next Fibonacci number in the sequence.
    """
    while True:
        yield base + scale * a
        a, b = b, a + b

# Instrument mapping (from https://raw.githubusercontent.com/FoxLisk/midifier/b8b276fe3ff9b8fe159b9dc4046c1d9f0e62ea29/midifier/instruments.py)
instrument_mapping = {
  'Acoustic Grand Piano': 0,
  'Bright Acoustic Piano': 1,
  'Electric Grand Piano': 2,
  'Honky-tonk Piano': 3,
  'Electric Piano 1': 4,
  'Electric Piano 2': 5,
  'Harpsichord': 6,
  'Clavinet': 7,
  'Celesta': 8,
  'Glockenspiel': 9,
  'Music Box': 10,
  'Vibraphone': 11,
  'Marimba': 12,
  'Xylophone': 13,
  'Tubular Bells': 14,
  'Dulcimer': 15,
  'Drawbar Organ': 16,
  'Percussive Organ': 17,
  'Rock Organ': 18,
  'Church Organ': 19,
  'Reed Organ': 20,
  'Accordion': 21,
  'Harmonica': 22,
  'Tango Accordion': 23,
  'Acoustic Guitar (nylon)': 24,
  'Acoustic Guitar (steel)': 25,
  'Electric Guitar (jazz)': 26,
  'Electric Guitar (clean)': 27,
  'Electric Guitar (muted)': 28,
  'Overdriven Guitar': 29,
  'Distortion Guitar': 30,
  'Guitar Harmonics': 31,
  'Acoustic Bass': 32,
  'Electric Bass (finger)': 33,
  'Electric Bass (pick)': 34,
  'Fretless Bass': 35,
  'Slap Bass 1': 36,
  'Slap Bass 2': 37,
  'Synth Bass 1': 38,
  'Synth Bass 2': 39,
  'Violin': 40,
  'Viola': 41,
  'Cello': 42,
  'Contrabass': 43,
  'Tremolo Strings': 44,
  'Pizzicato Strings': 45,
  'Orchestral Harp': 46,
  'Timpani': 47,
  'String Ensemble 1': 48,
  'String Ensemble 2': 49,
  'Synth Strings 1': 50,
  'Synth Strings 2': 51,
  'Choir Aahs': 52,
  'Voice Oohs': 53,
  'Synth Choir': 54,
  'Orchestra Hit': 55,
  'Trumpet': 56,
  'Trombone': 57,
  'Tuba': 58,
  'Muted Trumpet': 59,
  'French Horn': 60,
  'Brass Section': 61,
  'Synth Brass 1': 62,
  'Synth Brass 2': 63,
  'Soprano Sax': 64,
  'Alto Sax': 65,
  'Tenor Sax': 66,
  'Baritone Sax': 67,
  'Oboe': 68,
  'English Horn': 69,
  'Bassoon': 70,
  'Clarinet': 71,
  'Piccolo': 72,
  'Flute': 73,
  'Recorder': 74,
  'Pan Flute': 75,
  'Blown bottle': 76,
  'Shakuhachi': 77,
  'Whistle': 78,
  'Ocarina': 79,
  'Lead 1 (square)': 80,
  'Lead 2 (sawtooth)': 81,
  'Lead 3 (calliope)': 82,
  'Lead 4 (chiff)': 83,
  'Lead 5 (charang)': 84,
  'Lead 6 (voice)': 85,
  'Lead 7 (fifths)': 86,
  'Lead 8 (bass + lead)': 87,
  'Pad 1 (new age)': 88,
  'Pad 2 (warm)': 89,
  'Pad 3 (polysynth)': 90,
  'Pad 4 (choir)': 91,
  'Pad 5 (bowed)': 92,
  'Pad 6 (metallic)': 93,
  'Pad 7 (halo)': 94,
  'Pad 8 (sweep)': 95,
  'FX 1 (rain)': 96,
  'FX 2 (soundtrack)': 97,
  'FX 3 (crystal)': 98,
  'FX 4 (atmosphere)': 99,
  'FX 5 (brightness)': 100,
  'FX 6 (goblins)': 101,
  'FX 7 (echoes)': 102,
  'FX 8 (sci-fi)': 103,
  'Sitar': 104,
  'Banjo': 105,
  'Shamisen': 106,
  'Koto': 107,
  'Kalimba': 108,
  'Bagpipe': 109,
  'Fiddle': 110,
  'Shanai': 111,
  'Tinkle Bell': 112,
  'Agogo': 113,
  'Steel Drums': 114,
  'Woodblock': 115,
  'Taiko Drum': 116,
  'Melodic Tom': 117,
  'Synth Drum': 118,
  'Reverse Cymbal': 119,
  'Guitar Fret Noise': 120,
  'Breath Noise': 121,
  'Seashore': 122,
  'Bird Tweet': 123,
  'Telephone Ring': 124,
  'Helicopter': 125,
  'Applause': 126,
  'Gunshot': 127,
}


import random
import numpy as np


class Chain:
    """
    A class representing a chain of sequences generated using random walks.

    Attributes:
        walk_range (list): The range of step choices for the random walk. Defaults to [-1, 0, 1].
        walk_start (int): The starting position for the random walk. Defaults to the middle value of walk_range.
        walk_probability (list or object): The probability distribution for choosing a step in the random walk.
            Can be a list of step choices or a random variable object. Defaults to [-1, 0, 1].
        branching_probability (float): The probability of branching at each step. Defaults to 0.0.
        merging_probability (float): The probability of merging sequences with the same last value. Defaults to 0.0.
    """

    def __init__(self, walk_range=None, walk_start=None, walk_probability=None, round_to=None, branching_probability=0.0, merging_probability=0.0):
        self.walk_range = walk_range
        self.walk_start = walk_start if walk_start is not None else (self.walk_range[1] - self.walk_range[0]) // 2
        self.walk_probability = walk_probability or [-1, 0, 1]  # Default step choices
        self.branching_probability = branching_probability
        self.merging_probability = merging_probability
        self.round_to = round_to

    def generate(self, length, seed=None):
        """
        Generates a chain of sequences using random walks.

        Args:
            length (int): The length of each sequence in the chain. Defaults to 10.
            seed (int): The seed value for the random number generator. Defaults to None.

        Returns:
            list: A list of sequences generated using random walks.
        """
        self.length = length
        random.seed(seed)
        np.random.seed(seed)
        sequences = [[self.walk_start]]
        for _ in range(self.length - 1):
            new_sequences = []
            for seq in sequences:
                if seq[-1] is not None:  # Check if sequence is active
                    # Find the index of the last value in the scale
                    last_value = seq[-1]

                    if isinstance(self.walk_probability, list):
                        step = random.choice(self.walk_probability)  # Choose a random step
                    elif hasattr(self.walk_probability, 'rvs'):
                        step = self.walk_probability.rvs()  # Draw a sample and round to nearest integer
                    
                    if self.round_to is not None:
                        step = round(step, self.round_to)

                    next_value = last_value + step  # Calculate next index

                    # Handle boundary conditions for next_value
                    if self.walk_range is not None:
                        if next_value < self.walk_range[0]:
                            next_value = self.walk_range[0]
                        elif next_value > self.walk_range[1]:
                            next_value = self.walk_range[1]

                    # Branching decision
                    if random.random() < self.branching_probability:
                        # Create new branch with None values up to the current length
                        new_branch = [None for _ in range(len(seq))]  
                        new_branch.append(next_value)  # Start new branch from next value value
                        new_sequences.append(new_branch)  # Add new branch
                        seq.append(next_value)  # Continue current sequence with value value
                    else:
                        seq.append(next_value)  # Continue without branching with value value
                    
            # Handle merging
            if random.random() < self.merging_probability:
                unique_values = set(s[-1] for s in sequences if s[-1] is not None)
                for value in unique_values:
                    value_seqs = [s for s in sequences if s[-1] == value]
                    if len(value_seqs) > 1:
                        # Keep the longest sequence, close others
                        longest_seq = max(value_seqs, key=len)
                        for s in value_seqs:
                            if s != longest_seq:
                                s[len(s):] = [None] * (self.length - len(s))  # Extend with None without overwriting last value
            sequences.extend(new_sequences)  # Incorporate new branches
            
        # Ensure all sequences have proper length
        for seq in sequences:
            if len(seq) < self.length:
                seq.extend([None] * (self.length - len(seq)))  # Fill with None to reach desired length

        return sequences  # Return the updated sequences
    

class Kernel:
    """
    A class representing a kernel for generating sequences.

    Attributes:
        walk_around (float): The mean value around which the sequence will walk.
        data (ndarray): The input data used for generating the sequence.
        length_scale (float): The length scale parameter of the kernel.
        amplitude (float): The amplitude parameter of the kernel.
    """

    def __init__(self, walk_around=0.0, length_scale=1.0, amplitude=1.0):
        self.walk_around = walk_around
        self.length_scale = length_scale
        self.amplitude = amplitude

    def rbf_kernel(self, dimension, length_scale):
        """
        Computes the radial basis function (RBF) kernel matrix.

        Args:
            dimension (int): The dimension of the kernel matrix.
            length_scale (float): The length scale parameter of the kernel.

        Returns:
            ndarray: The RBF kernel matrix.
        """
        covariance_matrix = np.zeros((dimension, dimension))
        for i in range(dimension):
            for j in range(dimension):
                distance_squared = (i - j) ** 2
                covariance_matrix[i, j] = np.exp(-distance_squared / (2 * length_scale ** 2))
                
        return covariance_matrix

    def generate(self, length=10, data=None, nsamples=1, seed=None):
        """
        Generates a sequence using the kernel.

        Args:
            length (int): The length of the sequence.
            data (ndarray): The input data used for generating the sequence.
            nsamples (int): The number of samples to generate.
            seed (int): The seed value for random number generation.

        Returns:
            list: The generated sequence.
        """
        self.length = length
        if data is not None:
            if not isinstance(data, np.ndarray) or data.ndim != 2 or data.shape[1] != 2:
                raise ValueError("data must be a two-dimensional NumPy array with two columns.")
        if not (isinstance(self.length_scale, (float, int)) and self.length_scale > 0):
            raise ValueError("length_scale must be a positive int or float.")
        if not (isinstance(self.amplitude, (float, int)) and self.amplitude > 0):
            raise ValueError("amplitude must be a positive int or float.")
        if not (isinstance(nsamples, int) and nsamples > 0):
            raise ValueError("nsamples must be a positive integer.")
        
        self.data = data

        random.seed(seed)
        np.random.seed(seed)

        if self.data is None:
            kernel_cov = self.walk_around + self.amplitude * self.rbf_kernel(dimension = self.length, length_scale = self.length_scale)
            
            sequence = []
            for _ in range(nsamples):
                sequence.append(np.random.multivariate_normal(
                    mean=np.repeat(self.walk_around, self.length),
                    cov=kernel_cov
                    ).tolist()
                )
        else:
            try:
                from sklearn.gaussian_process import GaussianProcessRegressor
                from sklearn.gaussian_process.kernels import RBF
            except ImportError:
                raise ImportError("It seems that scikit-learn is not installed. Please install it using pip install scikit-learn or using your favorite installer.")
            x = np.linspace(0, self.data[:, 0].max(), self.length)[:, np.newaxis]
            kernel = self.amplitude * RBF(length_scale=self.length_scale) # , length_scale_bounds = (self.length_scale*0.9, self.length_scale*1.1)

            gp = GaussianProcessRegressor(kernel=kernel, normalize_y=True, n_restarts_optimizer=10)
            gp.fit(data[:, 0].reshape(-1, 1), data[:, 1].reshape(-1, 1))
            sequence = [
                x.flatten().tolist(),
                gp.sample_y(x, n_samples=nsamples).T.tolist()
            ]
        
        return sequence
    
class CelestialBody:
    """
    Represents a celestial body in space.

    Attributes:
        distance (float): The distance of the celestial body from its parent body.
        orbital_speed (float): The orbital speed of the celestial body.
        phase (float, optional): The phase of the celestial body's orbit (default is 0).
        moons (list, optional): A list of CelestialBody objects representing the moons of the celestial body (default is None).
    """

    def __init__(self, distance, orbital_speed, phase=0, moons=None):
        """
        Initializes a new instance of the CelestialBody class.

        Args:
            distance (float): The distance of the celestial body from its parent body.
            orbital_speed (float): The orbital speed of the celestial body.
            phase (float, optional): The phase of the celestial body's orbit (default is 0).
            moons (list, optional): A list of CelestialBody objects representing the moons of the celestial body (default is None).
        """
        self.distance = distance
        self.orbital_speed = orbital_speed
        self.phase = phase
        self.moons = moons if moons else []

    def position(self, times):
        """
        Calculates the position of the celestial body at the given times.

        Args:
            times (array-like): An array-like object containing the times at which to calculate the position.

        Returns:
            tuple: A tuple containing the x and y coordinates of the celestial body's position at the given times.
        """
        # Vectorize the position calculations
        angles = self.orbital_speed * times + self.phase
        x = self.distance * np.cos(angles)
        y = self.distance * np.sin(angles)
        return x, y

    def simulate(self, times, parent_position=None):
        """
        Simulates the motion of the celestial body over the given times.

        Args:
            times (array-like): An array-like object containing the times at which to simulate the motion.
            parent_position (tuple, optional): A tuple containing the x and y coordinates of the parent body's position (default is None).

        Returns:
            list: A list of tuples representing the distances and angles of the celestial body at the given times.
        """
        if parent_position is None:
            parent_position = (np.zeros_like(times), np.zeros_like(times))

        x, y = self.position(times)
        x_total = parent_position[0] + x
        y_total = parent_position[1] + y

        results = []
        for moon in self.moons:
            results.extend(moon.simulate(times, (x_total, y_total)))

        if not self.moons:
            distances = np.sqrt(x_total**2 + y_total**2)
            angles = np.arctan2(y_total, x_total) * 180 / np.pi  # Convert to degrees
            return [(distances, angles % 360)]  # Ensure angles are from 0 to 360
        return results

class SolarSystem:
    """
    Represents a solar system.

    Attributes:
        planets (list): A list of celestial bodies in the solar system.
    """

    def __init__(self):
        """
        Initializes a new instance of the SolarSystem class.
        """
        self.planets = []

    def add_planet(self, planet):
        """
        Adds a new planet to the solar system.

        Args:
            planet (CelestialBody): A CelestialBody object representing the planet to add.
        """
        self.planets.append(planet)

    def simulate(self, times):
        """
        Simulates the motion of all planets in the solar system.

        Args:
            times (list): A list of time points at which to simulate the motion.

        Returns:
            list: A list of simulation results for all planets.
        """
        results = []
        for planet in self.planets:
            results.extend(planet.simulate(times))
        return results


import numpy as np
from scipy.signal import correlate, find_peaks


class Index:
    """
    A class that performs various analysis on a list of values.

    Args:
        values (list): A list of numerical values.
        weights (list, optional): A list of weights corresponding to the values. Defaults to None.

    Attributes:
        values (list): A list of numerical values.
        weights (list): A list of weights corresponding to the values.

    """

    def __init__(self, values, weights=None): 
        if weights is None:
            self.weights = [1] * len(values)
        else:
            self.weights = weights
        
        # Pair values and weights to filter out None values jointly
        cleaned_data = [(v, w) for v, w in zip(values, self.weights) if v is not None]
        self.values, self.weights = zip(*cleaned_data) if cleaned_data else ([], [])

        # Initialize positions assuming sequential values
        self.positions = list(range(len(self.values)))

    def gini(self):
        """
        Calculates the Gini index of the values.

        Returns:
            float: The Gini index.
        """
        if not self.values or len(self.values) <= 1:
            return None
        values = [v * w for v, w in zip(self.values, self.weights)]
        sorted_values = sorted(values)
        n = len(sorted_values)
        cumulative_sum = sum((i+1) * val for i, val in enumerate(sorted_values))
        total_sum = sum(sorted_values)
        gini = (2 * cumulative_sum) / (n * total_sum) - (n + 1) / n
        return gini

    def balance(self):
        """
        Calculates the balance index of the values.

        Returns:
            float: The balance index.

        """
        if not self.values or len(self.values) <= 1:
            return None
        weighted_positions = sum(pos * weight for pos, weight in zip(self.positions, self.weights))
        total_weight = sum(self.weights)
        center_of_mass = weighted_positions / total_weight
        total_length_of_cycle = max(self.positions) + (self.weights[self.positions.index(max(self.positions))])
        ideal_center = total_length_of_cycle / 2
        balance_index = abs(center_of_mass - ideal_center) / ideal_center
        return balance_index

    def autocorrelation(self):
        """
        Calculates the autocorrelation of the values.

        Returns:
            list: The autocorrelation values.

        """
        n = len(self.values)
        result = correlate(self.values, self.values, mode='full', method='fft')
        result = result[n-1:] / np.array([n - abs(i) for i in range(n)])
        return result.tolist()

    def motif(self):
        """
        Calculates the motif score of the values.

        Returns:
            float: The motif score.

        """
        if not self.values:  # Check if the list is empty
            return 0  # Or return another appropriate default value
        autocorr = self.autocorrelation()
        peaks, _ = find_peaks(autocorr)
        if len(peaks) > 1:
            motif_lengths = np.diff(peaks)  # Distances between peaks as potential motif lengths
            motif_length = np.median(motif_lengths)  # Use the median as a common motif length
        else:
            return 0  # Return zero score if no motif length is identified
        motif_length = int(motif_length)
        motif_counts = {}
        for i in range(len(self.values) - motif_length + 1):
            motif = tuple(self.values[i:i + motif_length])
            motif_counts[motif] = motif_counts.get(motif, 0) + 1
        # Score based on the frequency of repeated motifs
        motif_score = sum(count for count in motif_counts.values() if count > 1) / len(self.values)
        return motif_score

    def dissonance(self, scale):
        """
        Calculates the dissonance of the values with respect to a scale.

        Args:
            scale (list): A list of values representing a musical scale.

        Returns:
            float: The dissonance.

        """
        if not self.values:  # Check if the list is empty
            return 0  # Or return another appropriate default value
        n = 0
        for v in self.values:
            if v not in scale:
                n += 1
        return n/len(self.values)

    def rhythmic(self, measure_length):
        """
        Calculates the rhythmic score of the values.

        Args:
            measure_length (float): The length of a measure.

        Returns:
            float: The rhythmic score.

        """
        scores = []
        current_measure_duration = 0.0
        for duration in self.values:
            current_measure_duration += duration
            if current_measure_duration > measure_length:
                # Here we consider overflow as a negative, so we reset for the next measure
                scores.append(0)  # Score of 0 for overflowed measure, adjust based on your scoring preference
                current_measure_duration = duration  # Start counting the new measure
            elif current_measure_duration == measure_length:
                scores.append(1)  # Perfect fit for this measure
                current_measure_duration = 0.0  # Reset for the next measure

        # Handle last measure if it doesn't reach full measure_length but no more durations are available
        if 0 < current_measure_duration <= measure_length:
            # The closer the last measure's total duration is to the measure_length, the better
            scores.append(current_measure_duration / measure_length)

        # Return the average score if there are scores, else return 0 (indicating no fit)
        return np.mean(scores) if scores else 0
    
    def fibonacci_index(self):
        """
        Calculates a Fibonacci index to evaluate how closely the sequence matches a Fibonacci sequence.

        Returns:
            float: The Fibonacci index, lower values indicate closer match to Fibonacci sequence.
        """
        if len(self.values) < 3:
            return float('inf')  # Not enough data to compute Fibonacci likeness

        # Calculate ratios of consecutive numbers
        ratios = [self.values[i] / self.values[i-1] for i in range(1, len(self.values)) if self.values[i-1] != 0]

        # Calculate how these ratios deviate from the golden ratio
        golden_ratio = (1 + np.sqrt(5)) / 2
        deviations = [abs(ratio - golden_ratio) for ratio in ratios]

        # Calculate an index as the average of these deviations
        fibonacci_index = sum(deviations) / len(deviations)
        return fibonacci_index


import random
import itertools
from . import utils
from . import harmony

class MusicTheoryConstants():
    """
    The Base class defines a set of musical scales, intervals, and notes.
    - scale_to_triad method returns the intervals for a triad based on the given scale intervals.
    - note_to_triad method converts a note to a triad based on the given scale intervals.
    """
    chromatic_scale = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
    scale_intervals = {
        'major': [0, 2, 4, 5, 7, 9, 11],  # Ionian
        'minor': [0, 2, 3, 5, 7, 8, 10],  # Aeolian
        'diminished': [0, 2, 3, 5, 6, 8, 9, 11],
        'major pentatonic': [0, 2, 4, 7, 9],
        'minor pentatonic': [0, 3, 5, 7, 10],
        'chromatic': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        'lydian': [0, 2, 4, 6, 7, 9, 11],
        'mixolydian': [0, 2, 4, 5, 7, 9, 10],
        'dorian': [0, 2, 3, 5, 7, 9, 10],
        'phrygian': [0, 1, 3, 5, 7, 8, 10],
        'locrian': [0, 1, 3, 5, 6, 8, 10],
        'harmonic minor': [0, 2, 3, 5, 7, 8, 11],
        'melodic minor ascending': [0, 2, 3, 5, 7, 9, 11],
        'melodic minor descending': [0, 2, 3, 5, 7, 8, 10],  # same as natural minor
    }
    intervals = {'P1': 0, 'm2': 1, 'M2': 2, 'm3': 3, 'M3': 4, 'P4': 5, 'P5': 7, 'm6': 8, 'M6': 9, 'm7': 10, 'M7': 11, 'P8': 12}

    @staticmethod
    def convert_flat_to_sharp(note):
        # Mapping of flat notes to their equivalent sharp notes
        flat_to_sharp = {
            'Bb': 'A#', 'Db': 'C#', 'Eb': 'D#', 'Gb': 'F#', 'Ab': 'G#',
            'B-': 'A#', 'D-': 'C#', 'E-': 'D#', 'G-': 'F#', 'A-': 'G#'
        }
        return flat_to_sharp.get(note, note) 

    @staticmethod
    def scale_to_triad(scale):
        """Returns the intervals for a triad based on the given scale intervals."""
        return [scale[i] for i in [0, 2, 4]]  # root, third, fifth

    
class Scale(MusicTheoryConstants):
    """
    Represents a musical scale.

    Args:
        tonic (str): The tonic note of the scale.
        mode (str or list): The type of scale. Defaults to 'major'. If a list is provided, it represents a custom scale.

    Raises:
        ValueError: If the tonic note is not a valid note or if the scale type is not a valid scale.

    Attributes:
        tonic (str): The tonic note of the scale.
        mode (str or list): The type of scale.
    """

    def __init__(self, tonic, mode='major'):
        if tonic not in self.chromatic_scale:
            tonic = self.convert_flat_to_sharp(tonic)
            if tonic not in self.chromatic_scale:
                raise ValueError(f"'{tonic}' is not a valid tonic note. Select one among '{self.chromatic_scale}'.")
        self.tonic = tonic

        if isinstance(mode, list):
            self.scale_intervals['custom'] = mode
            mode = 'custom'
        elif mode not in self.scale_intervals.keys():
            raise ValueError(f"'{mode}' is not a valid scale. Select one among '{self.scale_intervals.keys()}' or a list of half steps such as [0, 2, 4, 5, 7, 9, 11] for a major scale.")
            
        self.mode = mode

    def generate(self):
        """
        Generates the full range of the scale.
    
        Returns:
            list: A list of MIDI note numbers representing the full range of the scale.
        """
        tonic_note = self.chromatic_scale.index(self.tonic)
        scale = self.scale_intervals.get(self.mode, self.scale_intervals['major'])
    
        full_range_scale = []
        added_notes = set()  # Keep track of added notes
    
        for octave in range(11):
            for interval in scale:
                note = (tonic_note + interval) % 12 + octave * 12
                if note <= 127 and note not in added_notes:
                    full_range_scale.append(note)
                    added_notes.add(note)
    
        full_range_scale.sort()
        return full_range_scale



class Progression(MusicTheoryConstants):
    """A class representing a musical progression generator based on the circle of fifths (or any other interval)."""

    def __init__(self, tonic_pitch='C4', circle_of='P5', type='chords', radius=[3, 3, 1], weights=None):
        """
        Initialize a Progression object.

        Args:
            tonic_pitch (str): The tonic pitch of the progression. Defaults to 'C4'.
            circle_of (str): The interval to form the circle of fifths. Defaults to 'P5'.
            type (str): The type of progression to generate. Can be 'chords' or 'pitches'. Defaults to 'chords'.
            radius (list): A list defining the range for major, minor, and diminished chords. Defaults to [3, 3, 1].
            weights (list): The weights for selecting chord types. If not provided, the radius values will be used as weights.

        Raises:
            ValueError: If the circle_of value is not among the available intervals.
            ValueError: If the type value is not 'chords' or 'pitches'.
        """
        self.tonic_midi = utils.cde_to_midi(tonic_pitch)
        self.circle_of = circle_of
        self.type = type
        self.radius = radius
        self.weights = weights if weights else radius

        if self.circle_of not in self.intervals.keys():
            raise ValueError(f"Select a circle_of among {self.intervals.keys()}.")
        if self.type not in ['chords', 'pitches']:
            raise ValueError("Type must either be 'pitches' or 'chords'.")

    def compute_circle(self):
        """
        Compute chords based on the circle of fifths, thirds, etc., within the specified radius.

        Returns:
            tuple: A tuple containing lists of root MIDI notes for major, minor, and diminished chords.
        """
        n_semitones = self.intervals[self.circle_of]
        circle_notes = [self.tonic_midi]
        for _ in range(max(self.radius)):
            next_note = (circle_notes[-1] + n_semitones) % 12 + (circle_notes[-1] // 12) * 12
            circle_notes.append(next_note)

        major_roots = circle_notes[:self.radius[0]]
        minor_roots = circle_notes[:self.radius[1]]
        diminished_roots = circle_notes[:self.radius[2]]

        return major_roots, minor_roots, diminished_roots

    def generate_chord(self, root_note_midi, chord_type):
        """
        Generate a chord based on root MIDI note and chord type.

        Args:
            root_note_midi (int): The root MIDI note of the chord.
            chord_type (str): The type of chord to generate. Can be 'major', 'minor', or 'diminished'.

        Returns:
            list: A list of MIDI notes representing the generated chord.
        """
        chord_intervals = {
            'major': [0, 4, 7],
            'minor': [0, 3, 7],
            'diminished': [0, 3, 6]
        }

        intervals = chord_intervals.get(chord_type, [0, 4, 7])

        chord_notes = [(root_note_midi + interval) for interval in intervals]
        chord_notes = [note if note <= 127 else note - 12 for note in chord_notes]

        return chord_notes

    def generate(self, length=4, seed=None):
        """
        Generate a musical progression.

        Args:
            length (int): The length of the progression in number of chords. Defaults to 4.
            seed (int): The seed value for the random number generator. Defaults to None.

        Returns:
            list: A list of lists, where each inner list represents a chord in the progression.
        """
        if seed is not None:
            random.seed(seed)

        major_roots, minor_roots, diminished_roots = self.compute_circle()

        chord_roots = [major_roots, minor_roots, diminished_roots]
        progression = []
        for _ in range(length):
            chord_type_index = random.choices(range(3), weights=self.weights, k=1)[0]
            if chord_roots[chord_type_index]:
                root_note_midi = random.choice(chord_roots[chord_type_index])
                chord_type = ['major', 'minor', 'diminished'][chord_type_index]
                if isinstance(root_note_midi, list):
                    print('Warning: root_note_midi was a list, taking first element.')
                    root_note_midi = root_note_midi[0]
                chosen_chord = self.generate_chord(root_note_midi, chord_type)
                progression.append(chosen_chord)

        return progression


class Voice(MusicTheoryConstants):
    """
    A class to represent a musical voice.
    """  

    def __init__(self, mode='major', tonic='C', degrees=[0, 2, 4]):
        """
        Constructs all the necessary attributes for the voice object.

        Parameters
        ----------
            mode : str, optional
                The type of the scale (default is 'major').
            tonic : str, optional
                The tonic note of the scale (default is 'C').
            degrees : list, optional
                Relative degrees for chord formation (default is [0, 2, 4]).
        """
        self.tonic = tonic
        self.scale = harmony.Scale(tonic, mode).generate()  # a list of MIDI notes for the scale
        self.degrees = degrees  # relative degrees for chord formation

    def pitch_to_chord(self, pitch):
        """
        Convert a MIDI note to a chord based on the scale using the specified degrees.

        Parameters
        ----------
            pitch : int
                The MIDI note to convert.

        Returns
        -------
            list
                A list of MIDI notes representing the chord.
        """
        # to get the degree, I need a the tonic in the right octave, i.e. the tonic midi pitch
        octave = utils.get_octave(pitch)
        tonic_cde_pitch = self.tonic + str(octave)
        tonic_midi_pitch = utils.cde_to_midi(tonic_cde_pitch)

        # the degrees of the whole scale
        scale_degrees = [utils.get_degree_from_pitch(p, scale_list=self.scale, tonic_pitch=tonic_midi_pitch) for p in self.scale]
        pitch_degree = utils.get_degree_from_pitch(pitch, scale_list=self.scale, tonic_pitch=tonic_midi_pitch)
        pitch_degree = int(round(pitch_degree)) # round the degree if the pitch is out of scale

        chord = []
        for degree in self.degrees:
            absolute_degree = pitch_degree + degree
            absolute_index = scale_degrees.index(absolute_degree)
            chord.append(self.scale[absolute_index])
        return chord  # Chord is now directly from the scale

    def generate(self, notes, durations=None, arpeggios=False):
        """
        Generate chords or arpeggios based on the given notes.

        Args:
            notes (list or tuple): The notes to generate chords or arpeggios from.
            durations (list, optional): The durations of each note. If not provided, defaults to [1].
            arpeggios (bool, optional): If True, generate arpeggios instead of chords. Defaults to False.

        Returns:
            list: The generated chords or arpeggios.

        """
        
        if isinstance(notes, tuple):
            notes = [notes]
        if isinstance(notes[0], int): # if notes are in fact pitches
            if durations is None:
                durations = [1]
            durations_cycle = itertools.cycle(durations)
            current_offset = 0
            for i,p in enumerate(notes):
                d = next(durations_cycle)
                notes[i] = (p, d, current_offset)
                current_offset = current_offset + d
        
        chords = [(self.pitch_to_chord(p), d, o) for p, d, o in notes]
        
        if not arpeggios:
            return chords
        else:
            arpeggios_p = []
            for n in chords:
                pitches = n[0]
                for p in pitches:
                    arpeggios_p.append(p)
            arpeggios_n = []
            durations_cycle = itertools.cycle(durations) # reset cycle
            current_offset = 0
            for p in arpeggios_p:
                d = next(durations_cycle)
                arpeggios_n.append((p, d, current_offset))
                current_offset = current_offset + d
            return arpeggios_n



class Ornament(MusicTheoryConstants):

    def __init__(
            self, type='grace_note', tonic=None, mode=None, by=1.0,
            grace_note_type='acciaccatura', grace_pitches=None, trill_rate=0.125, arpeggio_degrees=None, slide_length=4.0
        ):
        """
        Initializes an Ornament object.

        Args:
            type (str): The type of ornament to be processed. Supported types include 'grace_note', 'trill', and 'mordent'.
            tonic (str): The tonic note for the scale.
            mode (str): The type of scale to generate.
            by (float): The pitch step for the trill.
            grace_note_type (str): Specifies the type of grace note ('acciaccatura' or 'appoggiatura') if applicable.
            grace_pitches (list): The list of pitches for the grace note.
            trill_rate (float): The duration of each individual note in the trill.
            arpeggio_degrees (list of integers): degrees in the scale to run the arpeggio
            slide_length (float): length of the slide
        """
        self.type = type
        if tonic and mode:
            self.tonic_index = self.chromatic_scale.index(tonic)  # Index in chromatic scale
            self.scale = self.generate_scale(tonic, mode)  # This will be a list of MIDI notes for the scale
            if arpeggio_degrees:
                self.arpeggio_voice = Voice(mode=mode, tonic=tonic, degrees=arpeggio_degrees)
            else:
                self.arpeggio_voice = None
        else:
            self.scale = None
            self.arpeggio_voice = None
        self.by = by
        self.grace_note_type = grace_note_type
        self.grace_pitches = grace_pitches
        self.trill_rate = trill_rate
        self.slide_length = slide_length
        

    def generate_scale(self, tonic, mode):
        """
        Generate a complete scale based on the tonic and scale type. This function is the same as the one in the Voice class.

        Args:
            tonic (str): The tonic note for the scale.
            mode (str): The type of scale to generate.
        
        Returns:
            list: A list of MIDI notes for the complete scale.
        """
        scale_pattern = self.scale_intervals[mode]
        scale_notes = [(self.tonic_index + interval) % 12 for interval in scale_pattern]  # Pitch classes
        complete_scale = []  # This will store the full scale in MIDI numbers
        for octave in range(-1, 10):  # Covering all MIDI octaves
            for note in scale_notes:
                midi_note = 12 * octave + note
                if 0 <= midi_note <= 127:  # Valid MIDI range
                    complete_scale.append(midi_note)
        return complete_scale

    def add_grace_note(self, notes, note_index):
        """
        Adds a grace note (either acciaccatura or appoggiatura) to a specified note in the list.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to which the trill will be added.

        Returns:
            list: The list of notes with the specified grace note added.
        """
        main_pitch, main_duration, main_offset = notes[note_index]
        ornament_pitch = random.choice(self.grace_pitches)
        if self.grace_note_type == 'acciaccatura':
            # Acciaccatura is very brief, does not alter the main note's start time.
            grace_duration = main_duration * 0.125  # Typically very short
            modified_main = (main_pitch, main_duration, main_offset + grace_duration)
            new_notes = notes[:note_index] + [(ornament_pitch, grace_duration, main_offset), modified_main] + notes[note_index + 1:]
        elif self.grace_note_type == 'appoggiatura':
            # Appoggiatura takes half the time of the main note and delays its start.
            grace_duration = main_duration / 2
            modified_main = (main_pitch, grace_duration, main_offset + grace_duration)
            new_notes = notes[:note_index] + [(ornament_pitch, grace_duration, main_offset), modified_main] + notes[note_index + 1:]
        else:
            # If neither, return the list unchanged
            new_notes = notes
        return new_notes

    def add_trill(self, notes, note_index):
        """
        Simulates a trill ornament by alternating between the original pitch and one step above.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to which the trill will be added.

        Returns:
            list: The list of notes with the specified trill applied to the specified note.
        """
        main_pitch, main_duration, main_offset = notes[note_index]
        trill_notes = []
        current_offset = main_offset
        
        # Determine the pitch to trill with based on the scale or semitone adjustment
        if self.scale and main_pitch in self.scale:
            pitch_index = self.scale.index(main_pitch)
            trill_pitch = self.scale[(pitch_index + int(round(self.by))) % len(self.scale)]  # Ensure the index wraps around the scale
        else:
            trill_pitch = main_pitch + self.by  # Default step if no scale is given or pitch is not in scale

        # Generate the sequence of trill notes to insert
        while current_offset < main_offset + main_duration:
            trill_notes.append((main_pitch, self.trill_rate, current_offset))
            trill_notes.append((trill_pitch, self.trill_rate, current_offset + self.trill_rate))
            current_offset += 2 * self.trill_rate

        # Insert the trill notes into the original list, replacing the original note at note_index
        new_notes = notes[:note_index] + trill_notes + notes[note_index + 1:]
        return new_notes


    def add_mordent(self, notes, note_index):
        """
        Simulates a mordent ornament by rapidly alternating between the original pitch and one step defined in `self.by`.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to which the trill will be added.

        Returns:
            list: A list containing the notes that make up the mordent.
        """
        main_pitch, main_duration, main_offset = notes[note_index]
        if self.scale and main_pitch in self.scale:
            pitch_index = self.scale.index(main_pitch)
            mordent_pitch = self.scale[pitch_index + int(round(self.by))]
        else:
            # If no scale is provided, default to moving a semitone down for a lower mordent
            # or a semitone up for an upper mordent
            mordent_pitch = main_pitch + self.by

        # The mordent splits the duration into three parts
        part_duration = main_duration / 3
        mordent_notes = [
            (main_pitch, part_duration, main_offset),
            (mordent_pitch, part_duration, main_offset + part_duration),
            (main_pitch, part_duration, main_offset + 2 * part_duration)
        ]
        new_notes = notes[:note_index] + mordent_notes + notes[note_index + 1:]
        return new_notes

    def add_arpeggiation(self, notes, note_index, voice):
        """
        Applies arpeggiation to a chord at a specified index in the list of notes using the degrees from a Voice instance.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to which the arpeggiation will be added.
            voice (Voice): An instance of the Voice class.

        Returns:
            list: The list of notes with the specified arpeggiation applied to the specified chord.
        """
        root_note = notes[note_index][0]
        main_duration, main_offset = notes[note_index][1], notes[note_index][2]
        
        # Generate the arpeggio notes based on the Voice's degrees
        arpeggio_notes = []
        for degree in voice.degrees:
            # Calculate pitch from the scale degree
            scale_degree_index = (voice.scale.index(root_note) + degree) % len(voice.scale)
            arpeggio_pitch = voice.scale[scale_degree_index]

            # Add an arpeggio note for each degree
            note_duration = main_duration / len(voice.degrees)
            note_offset = main_offset + voice.degrees.index(degree) * note_duration
            arpeggio_notes.append((arpeggio_pitch, note_duration, note_offset))

        # Replace the original note with the arpeggio notes
        new_notes = notes[:note_index] + arpeggio_notes + notes[note_index + 1:]
        return new_notes
    
    def add_turn(self, notes, note_index):
        """
        Simulates a turn ornament by playing the note above, the note itself, the note below, and returning to the note.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to which the turn will be added.

        Returns:
            list: The list of notes with the specified turn applied to the specified note.
        """
        main_pitch, main_duration, main_offset = notes[note_index]
        part_duration = main_duration / 4  # Splitting the total duration among the four notes of the turn

        if self.scale and main_pitch in self.scale:
            pitch_index = self.scale.index(main_pitch)
            upper_pitch = self.scale[(pitch_index + 1) % len(self.scale)]
            lower_pitch = self.scale[(pitch_index - 1 + len(self.scale)) % len(self.scale)]
        else:
            upper_pitch = main_pitch + self.by  # Assuming 'by' is the interval step
            lower_pitch = main_pitch - self.by

        turn_notes = [
            (upper_pitch, part_duration, main_offset),
            (main_pitch, part_duration, main_offset + part_duration),
            (lower_pitch, part_duration, main_offset + 2 * part_duration),
            (main_pitch, part_duration, main_offset + 3 * part_duration)
        ]
        
        new_notes = notes[:note_index] + turn_notes + notes[note_index + 1:]
        return new_notes

    def add_slide(self, notes, note_index, slide_length=4):
        """
        Simulates a slide from the current note to the next by incrementally changing the pitch.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note from which the slide will start.
            slide_length (int): The number of steps in the slide.

        Returns:
            list: The list of notes with the specified slide applied.
        """
        if note_index < len(notes) - 1:  # Ensure there is a following note to slide into
            start_pitch, _, start_offset = notes[note_index]
            end_pitch, end_duration, end_offset = notes[note_index + 1]

            # Calculate pitch steps for the slide, assuming a linear slide for simplicity
            pitch_step = (end_pitch - start_pitch) / slide_length
            slide_duration = (end_offset - start_offset) / slide_length
            
            slide_notes = [(int(start_pitch + pitch_step * i), slide_duration, start_offset + slide_duration * i) for i in range(slide_length)]
            
            # Insert the slide notes, remove the original note and the next one since they are now part of the slide
            new_notes = notes[:note_index] + slide_notes + notes[note_index + 2:]
        else:
            # No following note to slide into, so return the original notes unchanged
            new_notes = notes

        return new_notes


    def generate(self, notes, note_index=None):
        """
        Applies the specified ornamentation action and type to the list of notes.

        Args:
            notes (list): The list of notes to be processed.
            note_index (int): The index of the note to ornament. If None, a note will be chosen randomly.

        Returns:
            list: The list of notes with the specified ornamentation applied.
        """
        if note_index is None:
            note_index = random.randint(0, len(notes) - 1)
        
        if self.type == 'grace_note':
            return self.add_grace_note(notes, note_index)
        elif self.type == 'trill':
            return self.add_trill(notes, note_index)
        elif self.type == 'mordent':
            return self.add_mordent(notes, note_index)
        elif self.type == 'arpeggio':
            return self.add_arpeggiation(notes, note_index, self.arpeggio_voice)
        elif self.type == 'turn':
            return self.add_turn(notes, note_index)
        elif self.type == 'slide':
            return self.add_slide(notes, note_index, self.slide_length)
        else:
            return notes  # Return original if ornament type is not recognized or required data is missing




import anywidget
import traitlets
from typing import List, Union, Tuple
from .conversion import to_abc

class ABCjsWidget(anywidget.AnyWidget):
    abc = traitlets.Unicode().tag(sync=True)
    _esm = """
    import * as abcjs from "https://cdn.jsdelivr.net/npm/abcjs@6.2.2/+esm";
    export function render({ model, el }) {
        function updateABC() {
            el.innerHTML = '';
            abcjs.renderAbc(el, model.get('abc'));
        }
        model.on('change:abc', updateABC);
        updateABC();
    }
    export default { render };
    """

@staticmethod
def show(tracks: Union[List[Tuple[Union[int, List[int]], float, float]], List[List[Tuple[Union[int, List[int]], float, float]]]],
         key: str = "C",
         clef: str = "treble",
         time_signature: str = "4/4",
         title: str = None,
         tempo: int = 120
         ) -> 'ABCjsWidget':
    if isinstance(tracks[0], tuple):
        tracks = [tracks]
    abc = to_abc(tracks, key, clef, time_signature, title, tempo)
    widget = ABCjsWidget(abc=abc)
    return widget



import numpy as np
from . import utils

import plotly.graph_objects as go
import colorsys

class Polyloop:
    """
    Represents a collection of polyloops, which are sequences of musical notes.
    """

    def __init__(self, polyloops, measure_length=4, insert_rests=True):
        """
        Initializes a Polyloop object.

        Parameters:
        - polyloops (list): A list of polyloops. Each polyloop is expected to be in the form [(offset, pitch, duration), ...].
        - measure_length (int): The length of a measure in beats. Defaults to 4.
        - insert_rests (bool): Whether to insert rests in the polyloops. Defaults to True.
        """
        self.measure_length = measure_length
        self.polyloops = [utils.fill_gaps_with_rests(polyloop) for polyloop in polyloops] if insert_rests else polyloops

    def plot_polyloops(self, pulse=1/4, colors=None):
        """
        Plots the given polyloops as a radar chart, including arcs to represent the duration of each note.

        Parameters:
        - pulse (float): The duration of each pulse in beats. Defaults to 1/4.
        - colors (list): A list of colors to use for the plot. If not provided, a default color scheme will be used.

        Returns:
        - fig (plotly.graph_objects.Figure): The generated radar chart figure.
        """
        self.polyloops = [self.polyloops] if not any(isinstance(i, list) for i in self.polyloops) else self.polyloops
        polyloops_without_rests = [[note for note in polyloop if note[0] is not None] for polyloop in self.polyloops]

        n_polyloops = len(self.polyloops)
        traces = []
        #colors = go.Figure().layout.template.layout.colorway if colors is None else colors
        if colors is None:
            colors = [colorsys.hsv_to_rgb(i/n_polyloops, 1, 1) for i in range(n_polyloops)]
            #colors = [colorsys.hls_to_rgb(0, i/n_polyloops, 0) for i in range(n_polyloops)]
            colors = ['rgba(%d, %d, %d, 0.5)' % (int(r*255), int(g*255), int(b*255)) for r, g, b in colors]

        fig = go.Figure()

        for i, polyloop in enumerate(polyloops_without_rests):
            for _, duration, offset in polyloop:  # Ignore the pitch component
                start_theta, duration_theta = offset * 360 / self.measure_length, duration * 360 / self.measure_length
                arc = np.linspace(start_theta, start_theta + duration_theta, 100)  # Generate points for a smooth arc
                r = [n_polyloops-i-1] * 100  # Constant radius for the arc
                
                fig.add_trace(go.Scatterpolar(
                    r=r,
                    theta=arc % 360,  # Ensure theta is within 0-360 range
                    mode='lines',
                    line=dict(color='rgba(60, 60, 60, 0.65)', width=8), # colors[i % len(colors)]
                    name=f'Polyloop {i+1} Duration',
                    showlegend=False
                ))

            for _, duration, offset in polyloop:
                start_theta, end_theta = offset * 360 / self.measure_length, (offset + duration) * 360 / self.measure_length
                for theta in [start_theta, end_theta]:
                    fig.add_trace(go.Scatterpolar(
                        r=[n_polyloops-i-0.9, n_polyloops-i-1.1],
                        theta=[theta % 360, theta % 360],
                        mode='lines',
                        line=dict(color='Black', width=3),
                        name=f'Polyloop {i+1} Start/End',
                        showlegend=False
                    ))

            if polyloop:
                start_thetas = [offset * 360 / self.measure_length for _, _, offset in polyloop]
                start_thetas.append(start_thetas[0])

            traces.append(go.Scatterpolar(
                r=[n_polyloops-i-1]*(len(polyloop)+1),  # Account for the loop closure
                theta=start_thetas,
                mode='lines',
                line=dict(color='rgba(0, 0, 0, 0.65)', width=1),
                fill='toself',
                fillcolor=colors[i],
                name=f'Polyloop {i}',
                showlegend=False
            ))

        for trace in reversed(traces):
            fig.add_trace(trace)
        
        tickvals = np.linspace(0, 360, int(self.measure_length/pulse), endpoint=False)
        ticktext = [str(i % self.measure_length) for i in np.arange(0, self.measure_length, pulse)]
        radial_tickvals = np.arange(0, n_polyloops, 1)

        fig.update_layout(
            polar=dict(
                radialaxis=dict(
                    visible=True,
                    range=[n_polyloops, -0.1],
                    tickvals=radial_tickvals,
                    ticktext=[str(i) for i in radial_tickvals]
                ),
                angularaxis=dict(
                    tickvals=tickvals,
                    ticktext=ticktext,
                    direction="clockwise",
                    rotation=90
                )
            ),
            template='none',
            showlegend=True
        )

        fig.add_annotation(
            x=0.5, y=0.5, text="↻", showarrow=False, font=dict(size=30, color='White'),
            xref="paper", yref="paper"
        )

        return fig



import anywidget
import traitlets
from typing import List, Dict, Union, Tuple

class MusicPlayer(anywidget.AnyWidget):
    tracks = traitlets.List().tag(sync=True)
    custom_instruments = traitlets.Dict().tag(sync=True)
    initial_bpm = traitlets.Int(default_value=120).tag(sync=True)
    show_debug = traitlets.Bool(default_value=False).tag(sync=True)
    _esm = """
export async function render({ model, el }) {
    const colors = {
        background: '#FDF6E3',
        primary: '#333',
        secondary: '#EEE8D5',
        accent: '#333',
        text: '#073642',
        lightText: '#586E75'
    };

    const styles = {
        container: `
            font-family: 'Georgia', serif;
            background-color: ${colors.background};
            color: ${colors.text};
            padding: 20px;
            border-radius: 12px;
            width: 400px;
            box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
        `,
        controlsContainer: `
            display: flex;
            justify-content: space-between;
            align-items: center;
            margin-bottom: 15px;
        `,
        select: `
            padding: 8px;
            margin: 5px;
            border: none;
            border-radius: 6px;
            background-color: ${colors.secondary};
            color: ${colors.text};
            font-size: 14px;
            cursor: pointer;
            transition: background-color 0.3s ease;
        `,
        button: `
            flex: 0 0 80px;
            padding: 10px;
            border: none;
            border-radius: 6px;
            background-color: ${colors.primary};
            color: ${colors.background};
            font-size: 16px;
            cursor: pointer;
            transition: background-color 0.3s ease;
        `,
        bpmInput: `
            width: 60px;
            padding: 8px;
            margin: 0 10px;
            border: none;
            border-radius: 6px;
            background-color: ${colors.secondary};
            color: ${colors.text};
            font-size: 16px;
            text-align: center;
        `,
        timelineContainer: `
            position: relative;
            width: 100%;
            margin: 20px 0;
        `,
        timeline: `
            width: 100%;
            -webkit-appearance: none;
            background: ${colors.secondary};
            outline: none;
            border-radius: 15px;
            overflow: visible;
            height: 8px;
        `,
        timeDisplay: `
            display: flex;
            justify-content: space-between;
            font-size: 12px;
            color: ${colors.text};
            margin-top: 5px;
        `,
        buttonContainer: `
            display: flex;
            justify-content: space-between;
            margin-top: 15px;
        `,
        downloadButton: `
            padding: 10px 20px;
            margin: 0 5px;
            border: none;
            border-radius: 6px;
            background-color: ${colors.accent};
            color: ${colors.background};
            font-size: 14px;
            cursor: pointer;
            transition: background-color 0.3s ease;
            flex: 1;
        `
    };

    const createElementWithStyle = (tag, style) => {
        const element = document.createElement(tag);
        element.style.cssText = style;
        return element;
    };

    const container = createElementWithStyle('div', styles.container);
    const controlsContainer = createElementWithStyle('div', styles.controlsContainer);
    const synthSelectors = document.createElement('div');
    const bpmInput = createElementWithStyle('input', styles.bpmInput);
    Object.assign(bpmInput, { type: 'number', min: 60, max: 240, value: model.get('initial_bpm') });
    const playButton = createElementWithStyle('button', styles.button);
    playButton.textContent = '▶';
    const timelineContainer = createElementWithStyle('div', styles.timelineContainer);
    const timelineSlider = createElementWithStyle('input', styles.timeline);
    Object.assign(timelineSlider, { type: 'range', min: 0, max: 100, value: 0 });
    const timeDisplay = createElementWithStyle('div', styles.timeDisplay);
    const currentTime = document.createElement('span');
    currentTime.textContent = '0:00';
    const totalTime = document.createElement('span');
    totalTime.textContent = '0:00';
    timeDisplay.append(currentTime, totalTime);

    const showDebug = model.get('show_debug');
    const debugDiv = document.createElement('div');

    const updateDebug = msg => {
        if (showDebug) {
            model.set('debug_msg', msg);
            model.save_changes();
            debugDiv.textContent = msg;
        }
    };

    // Load Tone.js and @tonejs/midi dynamically
    await Promise.all([
        import('https://cdnjs.cloudflare.com/ajax/libs/tone/14.8.49/Tone.js'),
        import('https://unpkg.com/@tonejs/midi@2.0.28/build/Midi.js')
    ]);
    updateDebug('Tone.js and @tonejs/midi loaded');

    const customInstruments = model.get('custom_instruments');
    const synthTypes = ['Synth', 'AMSynth', 'FMSynth', 'MembraneSynth', ...Object.keys(customInstruments)];
    const synthSelectsArray = [];

    let musicData = model.get('tracks');
    musicData = Array.isArray(musicData[0]) ? musicData : [musicData];

    musicData.forEach((_, index) => {
        const synthSelectorItem = document.createElement('div');
        const synthLabel = document.createElement('label');
        synthLabel.textContent = `Instrument ${index + 1}`;
        const synthSelect = createElementWithStyle('select', styles.select);
        synthTypes.forEach(synth => {
            const option = document.createElement('option');
            option.value = synth;
            option.textContent = synth;
            synthSelect.appendChild(option);
        });
        synthSelectsArray.push(synthSelect);
        synthSelectorItem.append(synthLabel, synthSelect);
        synthSelectors.appendChild(synthSelectorItem);
    });

    let synths = [];
    let parts = [];
    let totalDuration;

    const formatTime = seconds => `${Math.floor(seconds / 60)}:${Math.floor(seconds % 60).toString().padStart(2, '0')}`;

    const createCustomInstrument = config => new Tone.Synth(config).toDestination();

    const initAudio = () => {
        synths.forEach(s => s.dispose());
        parts.forEach(p => p.dispose());
        synths = [];
        parts = [];

        musicData.forEach((instrumentData, index) => {
            const selectedSynth = synthSelectsArray[index].value;
            const synth = customInstruments[selectedSynth] 
                ? createCustomInstrument(customInstruments[selectedSynth].config)
                : new Tone[selectedSynth]().toDestination();
            synths.push(synth);

            const part = new Tone.Part((time, note) => {
                if (note.noteName === null) {
                    // C'est un silence, ne rien jouer
                    return;
                }
                if (Array.isArray(note.noteName)) {
                    // C'est un accord
                    note.noteName.forEach(n => {
                        if (n !== null) {
                            synth.triggerAttackRelease(n, note.duration, time);
                        }
                    });
                } else {
                    // C'est une note simple
                    synth.triggerAttackRelease(note.noteName, note.duration, time);
                }
            }, instrumentData.map(([midi, duration, time]) => ({
                time,
                noteName: midi === null ? null :
                    (Array.isArray(midi) 
                        ? midi.map(m => m !== null ? Tone.Frequency(m, "midi").toNote() : null)
                        : Tone.Frequency(midi, "midi").toNote()),
                duration
            }))).start(0);
            parts.push(part);
        });

        totalDuration = Math.max(...musicData.flat().map(note => note[2] + note[1]));
        Tone.Transport.loopEnd = totalDuration;
        Tone.Transport.loop = true;

        totalTime.textContent = formatTime(totalDuration);
        updateDebug(`Audio initialized, duration: ${totalDuration.toFixed(2)}s`);
    };

    const updateTimeline = () => {
        const progress = (Tone.Transport.seconds / totalDuration) * 100;
        timelineSlider.value = progress;
        currentTime.textContent = formatTime(Tone.Transport.seconds);
        if (Tone.Transport.state === 'started') {
            requestAnimationFrame(updateTimeline);
        }
    };

    playButton.onclick = async () => {
        try {
            if (Tone.Transport.state === 'started') {
                await Tone.Transport.stop();
                playButton.textContent = '▶';
                updateDebug('Playback stopped');
            } else {
                if (synths.length === 0) initAudio();
                await Tone.start();
                await Tone.Transport.start();
                playButton.textContent = '◼';
                updateDebug('Playback started');
                updateTimeline();
            }
        } catch (error) {
            updateDebug(`Error: ${error.message}`);
        }
    };

    synthSelectsArray.forEach(select => {
        select.onchange = initAudio;
    });

    timelineSlider.oninput = () => {
        const time = (timelineSlider.value / 100) * totalDuration;
        Tone.Transport.seconds = time;
        currentTime.textContent = formatTime(time);
    };

    bpmInput.onchange = () => {
        const bpm = parseInt(bpmInput.value);
        if (bpm >= 60 && bpm <= 240) {
            Tone.Transport.bpm.value = bpm;
            updateDebug(`BPM set to ${bpm}`);
        } else {
            bpmInput.value = Tone.Transport.bpm.value;
            updateDebug('Invalid BPM. Please enter a value between 60 and 240.');
        }
    };

    const buttonContainer = createElementWithStyle('div', styles.buttonContainer);

    const downloadMIDIButton = createElementWithStyle('button', styles.downloadButton);
    downloadMIDIButton.textContent = 'MIDI';

    const generateMIDI = () => {
        const midi = new Midi();
        const bpm = Tone.Transport.bpm.value;
        musicData.forEach(instrumentData => {
            const track = midi.addTrack();
            instrumentData.forEach(([midiNote, duration, time]) => {
                if (midiNote === null) {
                    // C'est un silence, ne rien ajouter à la piste MIDI
                    return;
                }
                if (Array.isArray(midiNote)) {
                    // C'est un accord
                    midiNote.forEach(note => {
                        if (note !== null) {
                            track.addNote({ midi: note, time, duration });
                        }
                    });
                } else {
                    // C'est une note simple
                    track.addNote({ midi: midiNote, time, duration });
                }
            });
        });
        midi.header.setTempo(bpm);
        return midi.toArray();
    };

    downloadMIDIButton.onclick = () => {
        const midiData = generateMIDI();
        const blob = new Blob([midiData], { type: "audio/midi" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'generated_music.mid';
        link.click();
        URL.revokeObjectURL(url);
        updateDebug('MIDI file downloaded');
    };

    const downloadWavButton = createElementWithStyle('button', styles.downloadButton);
    downloadWavButton.textContent = 'WAV';

    const audioBufferToWav = buffer => {
        const numOfChan = buffer.numberOfChannels;
        const length = buffer.length * numOfChan * 2 + 44;
        const out = new ArrayBuffer(length);
        const view = new DataView(out);
        let offset = 0;
        let pos = 0;

        const setUint16 = data => {
            view.setUint16(pos, data, true);
            pos += 2;
        };

        const setUint32 = data => {
            view.setUint32(pos, data, true);
            pos += 4;
        };

        setUint32(0x46464952);
        setUint32(length - 8);
        setUint32(0x45564157);
        setUint32(0x20746d66);
        setUint32(16);
        setUint16(1);
        setUint16(numOfChan);
        setUint32(buffer.sampleRate);
        setUint32(buffer.sampleRate * 2 * numOfChan);
        setUint16(numOfChan * 2);
        setUint16(16);
        setUint32(0x61746164);
        setUint32(length - pos - 4);

        const channels = Array.from({length: buffer.numberOfChannels}, (_, i) => buffer.getChannelData(i));

        while (pos < length) {
            channels.forEach(channel => {
                const sample = Math.max(-1, Math.min(1, channel[offset]));
                view.setInt16(pos, sample < 0 ? sample * 0x8000 : sample * 0x7FFF, true);
                pos += 2;
            });
            offset++;
        }

        return out;
    };

    const generateWAV = async () => {
        const duration = Math.max(...musicData.flat().map(note => note[2] + note[1]));

        const buffer = await Tone.Offline(({ transport }) => {
            musicData.forEach((instrumentData, index) => {
                const synthType = synthSelectsArray[index].value;
                const synth = customInstruments[synthType] 
                    ? createCustomInstrument(customInstruments[synthType].config)
                    : new Tone[synthType]().toDestination();

                new Tone.Part((time, note) => {
                    if (note.noteName === null) {
                        // C'est un silence, ne rien jouer
                        return;
                    }
                    if (Array.isArray(note.noteName)) {
                        // C'est un accord
                        note.noteName.forEach(n => {
                            if (n !== null) {
                                synth.triggerAttackRelease(n, note.duration, time);
                            }
                        });
                    } else {
                        // C'est une note simple
                        synth.triggerAttackRelease(note.noteName, note.duration, time);
                    }
                }, instrumentData.map(([midi, duration, time]) => ({
                    time,
                    noteName: midi === null ? null :
                        (Array.isArray(midi) 
                            ? midi.map(m => m !== null ? Tone.Frequency(m, "midi").toNote() : null)
                            : Tone.Frequency(midi, "midi").toNote()),
                    duration
                }))).start(0);
            });

            transport.start();
        }, duration);

        return new Blob([audioBufferToWav(buffer.get())], { type: 'audio/wav' });
    };

    downloadWavButton.onclick = async () => {
        updateDebug('Generating WAV file...');
        const wavBlob = await generateWAV();
        const url = URL.createObjectURL(wavBlob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'generated_music.wav';
        link.click();
        URL.revokeObjectURL(url);
        updateDebug('WAV file downloaded');
    };

    container.appendChild(synthSelectors);
    controlsContainer.append(bpmInput, playButton);
    container.appendChild(controlsContainer);
    timelineContainer.append(timelineSlider, timeDisplay);
    container.appendChild(timelineContainer);
    buttonContainer.append(downloadMIDIButton, downloadWavButton);
    container.appendChild(buttonContainer);

    if (showDebug) {
        container.appendChild(debugDiv);
    }
    el.appendChild(container);

    // Set initial BPM
    Tone.Transport.bpm.value = model.get('initial_bpm');
    updateDebug('Widget initialized');
}

export default { render };
    """

@staticmethod
def show(
    tracks: Union[List[Tuple[Union[int, List[int], None], float, float]], List[List[Tuple[Union[int, List[int], None], float, float]]]],
    custom_instruments: Dict[str, Dict] = None,
    initial_bpm: int = 120
) -> anywidget.AnyWidget:
    """
    Crée un lecteur de musique avancé.

    Args:
        tracks: Données musicales sous forme de liste de tuples (note MIDI, durée, temps) ou liste de listes de tuples.
        custom_instruments: Dictionnaire d'instruments personnalisés.
        initial_bpm: BPM initial.

    Returns:
        Un widget marimo contenant le lecteur de musique avancé.
    """
    if not isinstance(tracks[0], list):
        tracks = [tracks]

    custom_instruments = custom_instruments or {}

    return (
        MusicPlayer(
            tracks=tracks,
            custom_instruments=custom_instruments,
            initial_bpm=initial_bpm,
            debug_msg="Widget initialized",
        )
    )